MOCK_RESPONSE = {
    "openapi": "3.1.0",
    "info": {
        "title": "FrameX",
        "version": "0.2.3",
    },
    "paths": {
        "/proxy/mock/get": {
            "get": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Get",
                "operationId": "proxy_mock_get",
                "parameters": [
                    {
                        "name": "message",
                        "in": "query",
                        "required": True,
                        "schema": {
                            "type": "string",
                            "title": "Message",
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/proxy/mock/post": {
            "post": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Post",
                "operationId": "proxy_mock_post",
                "parameters": [
                    {
                        "name": "message",
                        "in": "query",
                        "required": True,
                        "schema": {
                            "type": "string",
                            "title": "Message",
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/proxy/mock/post_model": {
            "post": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Post Model",
                "operationId": "proxy_mock_post_model",
                "requestBody": {
                    "required": True,
                    "content": {"application/json": {"schema": {"$ref": "#/components/schemas/MockModel"}}},
                },
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/proxy/mock/black_get": {
            "get": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Get",
                "operationId": "proxy_mock_do_not_import",
                "parameters": [
                    {
                        "name": "message",
                        "in": "query",
                        "required": True,
                        "schema": {
                            "type": "string",
                            "title": "Message",
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/proxy/mock/black_post": {
            "post": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Post",
                "operationId": "proxy_mock_black_post",
                "parameters": [
                    {
                        "name": "message",
                        "in": "query",
                        "required": True,
                        "schema": {
                            "type": "string",
                            "title": "Message",
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/api/v1/proxy/mock/info": {
            "get": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Info",
                "operationId": "proxy_mock_info",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/proxy/mock/auth/get": {
            "get": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Get",
                "operationId": "proxy_mock_auth_get",
                "parameters": [
                    {
                        "name": "message",
                        "in": "query",
                        "required": True,
                        "schema": {
                            "type": "string",
                            "title": "Message",
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
        "/proxy/mock/auth/sget": {
            "get": {
                "tags": ["proxy"],
                "summary": "Proxy Mock Get",
                "operationId": "proxy_mock_auth_sget",
                "parameters": [
                    {
                        "name": "message",
                        "in": "query",
                        "required": True,
                        "schema": {
                            "type": "string",
                            "title": "Message",
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {"type": "object"}}},
                    }
                },
            }
        },
    },
    "components": {
        "schemas": {
            "MockModel": {
                "title": "MockModel",
                "type": "object",
                "required": ["id"],
                "properties": {
                    "id": {
                        "type": "integer",
                        "title": "Id",
                    },
                    "name": {
                        "type": "string",
                        "title": "Name",
                        "default": "default",
                    },
                },
            }
        }
    },
}



import pytest
from fastapi.testclient import TestClient

from framex.consts import API_STR
from framex.utils import cache_decode, cache_encode
from tests.test_plugins import ExchangeModel, SubModel


def test_get_proxy_version(client: TestClient):
    res = client.get(f"{API_STR}/proxy/mock/info").json()
    assert res == {"info": "i_am_mock_proxy_info"}


def test_get_proxy_get(client: TestClient):
    params = {"message": "hello world"}
    res = client.get("/proxy/mock/get", params=params).json()
    assert res == {"method": "GET", "params": params}


def test_get_proxy_post(client: TestClient):
    params = {"message": "hello world"}
    res = client.post("/proxy/mock/post", params=params).json()
    assert res == {"method": "POST", "params": params}


def test_get_proxy_post_model(client: TestClient):
    data = {"id": 12345, "name": "01234567890"}
    res = client.post("/proxy/mock/post_model", json=data).json()
    assert res == {"method": "POST", "body": data}


def test_get_proxy_black_get(client: TestClient):
    res = client.get("/proxy/mock/black_get").json()
    assert res["status"] == 404


def test_get_proxy_auth_get(client: TestClient):
    params = {"message": "hello world"}
    res = client.get("/proxy/mock/auth/get", params=params).json()
    assert res == {"method": "GET", "params": params}


def test_get_proxy_auth_sget(client: TestClient):
    params = {"message": "hello world"}
    res = client.get("/proxy/mock/auth/sget", params=params).json()
    assert res == {"method": "GET", "params": params}


@pytest.mark.order(2)
def test_call_proxy_func(client: TestClient):
    func = cache_encode("tests.test_plugins.local_exchange_key_value")
    data = cache_encode(
        {
            "a_str": "test",
            "b_int": 123,
            "c_model": ExchangeModel(id="id_1", name=100, model=SubModel(id=1, name="sub_name")),
        }
    )
    body = {"func_name": func, "data": data}
    headers = {"Authorization": "i_am_local_proxy_auth_keys"}
    res = client.post("/api/v1/proxy/remote", json=body, headers=headers).json()
    assert res["status"] == 200
    data = cache_decode(res["data"])
    assert data["a_str"] == "test"
    assert data["b_int"] == 123
    model = data["c_model"]
    assert model.id == "id_1"
    assert model.name == 100
    assert model.model.id == 1
    assert model.model.name == "sub_name"


@pytest.mark.order(2)
def test_call_proxy_func_with_error_func_name(client: TestClient):
    func = cache_encode("tests.test_plugins.error_func")
    data = cache_encode(
        {
            "a_str": "test",
            "b_int": 123,
            "c_model": ExchangeModel(id="id_1", name=100, model=SubModel(id=1, name="sub_name")),
        }
    )
    body = {"func_name": func, "data": data}
    headers = {"Authorization": "i_am_local_proxy_auth_keys"}
    res = client.post("/api/v1/proxy/remote", json=body, headers=headers).json()
    assert res["status"] == 500


@pytest.mark.order(2)
def test_call_proxy_func_with_error_api_key(client: TestClient):
    func = cache_encode("tests.test_plugins.local_exchange_key_value")
    data = cache_encode(
        {
            "a_str": "test",
            "b_int": 123,
            "c_model": ExchangeModel(id="id_1", name=100, model=SubModel(id=1, name="sub_name")),
        }
    )
    body = {"func_name": func, "data": data}
    headers = {"Authorization": "i_am_error_keys"}
    res = client.post("/api/v1/proxy/remote", json=body, headers=headers).json()
    assert res["status"] == 401



PROJECT_NAME = "FrameX"
VERSION = "0.2.12"
API_STR = "/api/v1"
OPENAPI_URL = "/api/v1/openapi.json"
DOCS_URL = "/docs"
REDOC_URL = "/redoc"

BACKEND_NAME = "backend"
APP_NAME = "default"

PROXY_PLUGIN_NAME = "proxy.ProxyPlugin"
PROXY_FUNC_HTTP_PATH = f"{API_STR}/proxy/remote"

DEFAULT_ENV = {"RAY_COLOR_PREFIX": "1", "RAY_DEDUP_LOGS": "1", "RAY_SERVE_RUN_SYNC_IN_THREADPOOL": "1"}


SEBTRY_BLOCK_URLS = [
    "/health",
    "/ping",
]



from collections.abc import Generator
from pathlib import Path
from unittest.mock import patch

import pytest
from click.testing import CliRunner
from fastapi import FastAPI
from fastapi.testclient import TestClient

import framex
from framex.config import settings
from tests.mock import mock_get, mock_request


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "record_mode": "new_episodes",  # new_episodes
        "filter_headers": ["authorization", "api-key"],
        "ignore_hosts": ["testserver"],
        "match_on": ["uri", "method", "body", "path", "query"],
        "allow_playback_repeats": True,
    }


@pytest.fixture(scope="module")
def disable_recording(request):
    return settings.test.disable_record_request


def before_record_request(request):
    if all(
        ch not in request.path
        for ch in [
            "rerank",
            "minio",
        ]
    ):
        return request
    return None


def before_record_response(response):
    if response["status"]["code"] != 200:
        return None
    return response


@pytest.fixture(scope="session", autouse=True)
def test_app() -> Generator:
    plugins = framex.load_plugins(str(Path(__file__).parent / "plugins"))
    assert len(plugins) == len(["invoker", "export", "alias_model"])

    with (
        patch("httpx.AsyncClient.get", new=mock_get),
        patch("httpx.AsyncClient.request", new=mock_request),
    ):
        yield framex.run(test_mode=True)  # type: ignore[return-value]


@pytest.fixture(scope="session")
def client(test_app: FastAPI) -> Generator:
    with TestClient(test_app, raise_server_exceptions=False) as c:
        yield c


@pytest.fixture
def runner():
    """Provide a reusable Click CLI runner."""
    return CliRunner()



import importlib
from collections.abc import Callable
from functools import wraps
from typing import Any, Literal, Optional, Union, get_args, get_origin

import pytest
from pydantic import BaseModel, ConfigDict, Field

from framex.plugin.load import register_proxy_func
from framex.plugin.on import on_proxy


class MatchMataConfig(BaseModel):
    disable_duplicate_school_detect: bool = Field(False, alias="disableDuplicateSchoolDetect")
    enable_talent_find: bool = Field(False, alias="enableTalentFind")
    use_fast: bool = Field(True, alias="useFast")
    max_team_member: int = Field(2, alias="maxTeamMember")
    cache: dict[str, str] = Field({})
    disable_cache: bool = Field(True, alias="disableCache")
    enable_random_sort_team: bool = Field(False, alias="enableRandomSortTeam")
    user_star_display: bool = Field(True, alias="userStarDisplay")
    ignore_teacher_score: bool = Field(False, alias="ignoreTeacherScore")
    algorithm_version: Literal["v3", "anchor"] = Field("anchor", alias="algorithmVersion")
    advanced_match: bool = Field(False, alias="advancedMatch")
    is_search_all: bool = Field(False, alias="isSearchAll")
    enable_matching_for_enterprise: bool = Field(False, alias="enableMatchingForEnterprise")

    model_config = ConfigDict(populate_by_name=True)


def resolve_base_type(annotation: Any) -> Any:
    origin = get_origin(annotation)
    if origin is Union:
        args = get_args(annotation)
        for arg in args:
            if arg is not type(None):
                return resolve_base_type(arg)
    if origin is Literal:
        return str
    return origin or annotation


def get_alias_set(fields: dict[str, Any]) -> set[str]:
    return {field.alias or name for name, field in fields.items()}


def get_alias_type_map(fields: dict[str, Any]) -> dict[str, type]:
    return {(field.alias or name): resolve_base_type(field.annotation) for name, field in fields.items()}


def test_resolve_annotation():
    builder = importlib.import_module("framex.plugins.proxy.builder")
    resolve_annotation: Callable = getattr(builder, "resolve_annotation")
    create_pydantic_model: Callable = getattr(builder, "create_pydantic_model")
    reset_created_models: Callable = getattr(builder, "reset_created_models")
    reset_created_models()

    # Test BaseType gen
    assert resolve_annotation({"type": "string", "title": "Task"}, {}) is str
    assert (
        resolve_annotation({"items": {"type": "string"}, "type": "array", "title": "Sub Tasks", "default": []}, {})
        == list[str]
    )
    assert (
        resolve_annotation({"items": {"type": "number"}, "type": "array", "title": "Weights", "default": []}, {})
        == list[float]
    )

    # Test BaseModel gen
    prop_schema: dict[str, Any] = {"$ref": "#/components/schemas/MatchMataConfig"}
    components = {
        "MatchMataConfig": {
            "properties": {
                "disableDuplicateSchoolDetect": {
                    "type": "boolean",
                    "title": "Disableduplicateschooldetect",
                    "default": False,
                },
                "enableTalentFind": {"type": "boolean", "title": "Enabletalentfind", "default": False},
                "useFast": {"type": "boolean", "title": "Usefast", "default": True},
                "maxTeamMember": {"type": "integer", "title": "Maxteammember", "default": 2},
                "cache": {
                    "additionalProperties": {"type": "string"},
                    "type": "object",
                    "title": "Cache",
                    "default": {},
                },
                "disableCache": {"type": "boolean", "title": "Disablecache", "default": True},
                "enableRandomSortTeam": {"type": "boolean", "title": "Enablerandomsortteam", "default": False},
                "userStarDisplay": {"type": "boolean", "title": "Userstardisplay", "default": True},
                "ignoreTeacherScore": {"type": "boolean", "title": "Ignoreteacherscore", "default": False},
                "algorithmVersion": {
                    "type": "string",
                    "enum": ["v3", "anchor"],
                    "title": "Algorithmversion",
                    "default": "anchor",
                },
                "advancedMatch": {"type": "boolean", "title": "Advancedmatch", "default": False},
                "isSearchAll": {"type": "boolean", "title": "Issearchall", "default": False},
                "enableMatchingForEnterprise": {
                    "type": "boolean",
                    "title": "Enablematchingforenterprise",
                    "default": False,
                },
            },
            "type": "object",
            "title": "MatchMataConfig",
        }
    }
    f1 = resolve_annotation(prop_schema, components).model_fields
    f2 = MatchMataConfig.model_fields

    assert get_alias_type_map(f1) == get_alias_type_map(f2), "Alias sets do not match"

    assert (
        resolve_annotation(
            {"anyOf": [{"items": {"type": "string"}, "type": "array"}, {"type": "string"}], "title": "Values"}, {}
        )
        == list[str] | str
    )
    assert (
        str(
            resolve_annotation(
                {"type": "array", "items": {"$ref": "#/components/schemas/MatchMataConfig"}, "title": "Values"},
                components,
            )
        )
        == "list[framex.plugins.proxy.builder.MatchMataConfig]"
    )

    # Test Exception
    with pytest.raises(RuntimeError) as exc_info:
        resolve_annotation({"$ref": "#/components/schemas/MatchMataConfig"}, {})

    assert "Unsupported prop_schema" in str(exc_info.value)

    # Test BaseModel gen
    model_schema = {
        "properties": {
            "dimensions": {"$ref": "#/components/schemas/Dimensions"},
            "global_filters": {"anyOf": [{"$ref": "#/components/schemas/GlobalFilters"}, {"type": "null"}]},
            "top_k": {"anyOf": [{"type": "integer"}, {"type": "null"}], "title": "Top K"},
        },
        "type": "object",
        "required": ["dimensions"],
        "title": "FetchCompanyByDimInfoRequest",
    }
    components = {
        "DimensionItem": {
            "properties": {
                "query_vector_text": {"type": "string", "title": "Query Vector Text"},
                "weight": {"type": "number", "title": "Weight"},
            },
            "type": "object",
            "required": ["query_vector_text", "weight"],
            "title": "DimensionItem",
        },
        "Dimensions": {
            "properties": {
                "dim_products": {"anyOf": [{"$ref": "#/components/schemas/DimensionItem"}, {"type": "null"}]},
                "dim_partners": {"anyOf": [{"$ref": "#/components/schemas/DimensionItem"}, {"type": "null"}]},
            },
            "type": "object",
            "title": "Dimensions",
        },
        "FetchCompanyByDimInfoRequest": {
            "properties": {
                "dimensions": {"$ref": "#/components/schemas/Dimensions"},
                "global_filters": {"anyOf": [{"$ref": "#/components/schemas/GlobalFilters"}, {"type": "null"}]},
                "top_k": {"anyOf": [{"type": "integer"}, {"type": "null"}], "title": "Top K"},
            },
            "type": "object",
            "required": ["dimensions"],
            "title": "FetchCompanyByDimInfoRequest",
        },
        "GlobalFilters": {
            "properties": {
                "provinces": {
                    "anyOf": [{"items": {"type": "string"}, "type": "array"}, {"type": "string"}, {"type": "null"}],
                    "title": "Provinces",
                },
                "cities": {
                    "anyOf": [{"items": {"type": "string"}, "type": "array"}, {"type": "string"}, {"type": "null"}],
                    "title": "Cities",
                },
                "is_listed": {"anyOf": [{"type": "integer"}, {"type": "null"}], "title": "Is Listed"},
            },
            "type": "object",
            "title": "GlobalFilters",
        },
        "HTTPValidationError": {
            "properties": {
                "detail": {
                    "items": {"$ref": "#/components/schemas/ValidationError"},
                    "type": "array",
                    "title": "Detail",
                }
            },
            "type": "object",
            "title": "HTTPValidationError",
        },
        "ValidationError": {
            "properties": {
                "loc": {
                    "items": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
                    "type": "array",
                    "title": "Location",
                },
                "msg": {"type": "string", "title": "Message"},
                "type": {"type": "string", "title": "Error Type"},
            },
            "type": "object",
            "required": ["loc", "msg", "type"],
            "title": "ValidationError",
        },
    }
    model = create_pydantic_model("FetchCompanyByDimInfoRequest", model_schema, components)
    fields = model.model_fields
    assert set(fields.keys()) == {"dimensions", "global_filters", "top_k"}
    assert fields["dimensions"].is_required() is True
    assert fields["global_filters"].default is None
    assert fields["top_k"].default is None
    # Check annotation types
    assert fields["top_k"].annotation == int | None


def test_resolve_default():
    builder = importlib.import_module("framex.plugins.proxy.builder")
    resolve_default: Callable = getattr(builder, "resolve_default")

    assert resolve_default(int) == 0
    assert resolve_default(str) == ""
    assert resolve_default(list[int]) == []
    assert resolve_default(dict[str, int]) == {}
    assert resolve_default(set[int]) == set()
    assert resolve_default(Union[int, str]) == 0  # noqa
    assert resolve_default(Optional[int]) is None  # noqa
    assert resolve_default(MatchMataConfig) == MatchMataConfig()

    with pytest.raises(RuntimeError) as exc_info:
        resolve_default(lambda x: x)
    assert "Cannot instantiate default" in str(exc_info.value)


class SubModel(BaseModel):
    id: int
    name: str


class ExchangeModel(BaseModel):
    id: str
    name: int
    model: SubModel


def supply_exception(func):
    @wraps(func)
    async def wrapper(*_: Any, **__: Any) -> None:
        raise RuntimeError("I am def supply_exception(func): exception")

    return wrapper


@on_proxy()
async def local_exchange_key_value(a_str: str, b_int: int, c_model: ExchangeModel) -> Any:
    return {"a_str": a_str, "b_int": b_int, "c_model": c_model}


@on_proxy()
@supply_exception
async def remote_exchange_key_value(a_str: str, b_int: int, c_model: ExchangeModel) -> Any:
    raise RuntimeError("This function should be called remotely")


@pytest.mark.order(1)
async def test_on_proxy_local_call():
    await register_proxy_func(local_exchange_key_value)
    res = await local_exchange_key_value(
        a_str="test", b_int=123, c_model=ExchangeModel(id="id_1", name=100, model=SubModel(id=1, name="sub_name"))
    )
    assert res["a_str"] == "test"
    assert res["b_int"] == 123
    model = res["c_model"]
    assert model.id == "id_1"
    assert model.name == 100
    assert model.model.id == 1
    assert model.model.name == "sub_name"


async def test_on_proxy_remote_call():
    res = await remote_exchange_key_value(
        a_str="test", b_int=123, c_model=ExchangeModel(id="id_1", name=100, model=SubModel(id=1, name="sub_name"))
    )
    assert res["result"] == "tests.test_plugins.remote_exchange_key_value"
    assert res["data"] == {
        "a_str": "test",
        "b_int": 123,
        "c_model": ExchangeModel(id="id_1", name=100, model=SubModel(id=1, name="sub_name")),
    }



from framex.config import OauthConfig


def test_config():
    from framex.plugin import get_plugin_config
    from framex.plugins.proxy.config import ProxyPluginConfig

    cfg = get_plugin_config("proxy", ProxyPluginConfig)
    assert isinstance(cfg, ProxyPluginConfig)
    assert cfg.proxy_urls is not None


def test_oauth_config_callback_url_property():
    cfg = OauthConfig(
        app_url="https://example.com",
        redirect_uri="/auth/callback",
    )
    assert cfg.call_back_url == "https://example.com/auth/callback"


def test_oauth_config_generates_default_urls_from_base_url():
    cfg = OauthConfig(
        base_url="https://gitlab.example.com",
    )

    assert cfg.authorization_url == "https://gitlab.example.com/oauth/authorize"
    assert cfg.token_url == "https://gitlab.example.com/oauth/token"  # noqa: S105
    assert cfg.user_info_url == "https://gitlab.example.com/api/v4/user"


def test_oauth_config_generates_jwt_secret_when_missing():
    cfg = OauthConfig()
    assert cfg.jwt_secret
    assert isinstance(cfg.jwt_secret, str)
    assert len(cfg.jwt_secret) >= 32


def test_oauth_config_does_not_override_custom_urls():
    cfg = OauthConfig(
        base_url="https://gitlab.example.com",
        authorization_url="https://custom.auth",
        token_url="https://custom.token",  # noqa: S106
        user_info_url="https://custom.user",
    )

    assert cfg.authorization_url == "https://custom.auth"
    assert cfg.token_url == "https://custom.token"  # noqa: S105
    assert cfg.user_info_url == "https://custom.user"



from fastapi import FastAPI

from framex.config import settings
from framex.consts import VERSION
from framex.log import LoguruHandler, logger


def _setup_env() -> None:
    import os

    from framex.consts import DEFAULT_ENV

    for key, value in DEFAULT_ENV.items():
        os.environ.setdefault(key, value)


def _setup_sentry(reversion: str | None = None) -> None:  # pragma: no cover
    if settings.sentry.enable and settings.sentry.dsn and settings.sentry.env:
        import sentry_sdk

        from framex.consts import SEBTRY_BLOCK_URLS

        def before_send(event, hint):  # noqa
            request = event.get("request")
            if not request:
                return event

            url = request.get("url", "")

            if any(path in url for path in SEBTRY_BLOCK_URLS):
                return None
            return event

        sentry_sdk.set_level("info")
        sentry_sdk.utils.logger.handlers = [LoguruHandler()]

        from framex.adapter import get_adapter

        adapter_mode = get_adapter().mode

        import os

        reversion = reversion or os.getenv("REVERSION")

        if reversion and not reversion.startswith("v"):
            reversion = f"v{reversion}"

        sentry_sdk.init(
            dsn=settings.sentry.dsn,
            debug=settings.sentry.debug,
            release=reversion,
            environment=f"{settings.sentry.env}_{adapter_mode}",
            ignore_errors=settings.sentry.ignore_errors,
            include_local_variables=True,
            send_default_pii=True,
            traces_sample_rate=1.0,
            profile_session_sample_rate=1.0,
            profile_lifecycle=settings.sentry.lifecycle,
            _experiments={
                "enable_logs": settings.sentry.enable_logs,
            },
            before_send=before_send,
        )
        logger.success(
            f"Successfully setup sentry with env: {settings.sentry.env}_{adapter_mode}, release: {reversion}"
        )


def run(
    *,
    server_host: str | None = None,
    server_port: int | None = None,
    dashboard_host: str | None = None,
    dashboard_port: int | None = None,
    num_cpus: int | None = None,
    use_ray: bool | None = None,
    enable_proxy: bool | None = None,
    load_builtin_plugins: list[str] | None = None,
    load_plugins: list[str] | None = None,
    reversion: str | None = None,
    blocking: bool = True,
    test_mode: bool = False,
) -> FastAPI | None:
    server_host = server_host if server_host is not None else settings.server.host
    server_port = server_port if server_port is not None else settings.server.port
    dashboard_host = dashboard_host if dashboard_host is not None else settings.server.dashboard_host
    dashboard_port = dashboard_port if dashboard_port is not None else settings.server.dashboard_port
    num_cpus = num_cpus if num_cpus is not None else settings.server.num_cpus
    use_ray = use_ray if use_ray is not None else settings.server.use_ray
    enable_proxy = enable_proxy if enable_proxy is not None else settings.server.enable_proxy
    builtin_plugins = settings.load_builtin_plugins if load_builtin_plugins is None else load_builtin_plugins
    external_plugins = settings.load_plugins if load_plugins is None else load_plugins

    if reversion:
        settings.server.reversion = reversion
    elif settings.server.reversion:
        reversion = settings.server.reversion
    else:
        reversion = VERSION
        settings.server.reversion = VERSION

    if test_mode and use_ray:
        raise RuntimeError("FlameX can not run when `test_mode` == True, and `use_ray` == True")

    # step1: setup log
    import sys

    from framex.log import StderrFilter

    # logging.basicConfig(handlers=[LoguruHandler()], level=0, force=True)
    sys.stderr = StderrFilter(sys.stderr, "file_system_monitor.cc:116:")

    # step2: setup env
    _setup_env()

    # step 4: setup settings plugins
    # Get all builtin_plugins

    from framex.plugin.load import auto_load_plugins

    auto_load_plugins(builtin_plugins, external_plugins, enable_proxy)

    # step5: init all DeploymentHandle
    logger.info("Start initializing all DeploymentHandle...")
    from framex.plugin import get_http_plugin_apis, init_all_deployments

    deployments = init_all_deployments(enable_proxy=enable_proxy)
    http_apis = get_http_plugin_apis()

    if use_ray:
        # step4: init ray
        try:
            import ray  # type: ignore[import-not-found]
            from ray import serve  # type: ignore[import-not-found]
        except ImportError as e:
            raise RuntimeError(
                'Ray engine requires extra dependency.\nInstall with: uv pip install "framex-kit[ray]"'
            ) from e

        ray.init(
            num_cpus=num_cpus if num_cpus > 0 else None,
            dashboard_host=dashboard_host,
            dashboard_port=dashboard_port,
            configure_logging=False,
            runtime_env={
                "env_vars": {
                    "REVERSION": reversion,
                },
                "worker_process_setup_hook": _setup_sentry,
            },
        )
        serve.start(
            detached=True,
            http_options={"host": server_host, "port": server_port},
        )
        from framex.driver.ingress import APIIngress

        api_ingress = APIIngress.bind(deployments=deployments, plugin_apis=http_apis)  # type: ignore

        serve.run(
            api_ingress,  # type: ignore
            blocking=blocking,
        )
    else:
        _setup_sentry(reversion=reversion)

        import uvicorn

        from framex.driver.ingress import APIIngress, app

        api_ingress = APIIngress(deployments=deployments, plugin_apis=http_apis)

        if test_mode:
            return app

        from framex.log import LOGGING_CONFIG

        uvicorn.run(  # pragma: no cover
            app,
            host=server_host,
            port=server_port,
            reload=False,
            loop="asyncio",
            log_config=LOGGING_CONFIG,
        )

    return None  # pragma: no cover


from framex.plugin import (
    BasePlugin,
    PluginApi,
    PluginMetadata,
    get_plugin,
    get_plugin_config,
    load_builtin_plugins,
    load_plugins,
    on_proxy,
    on_register,
    on_request,
    register_proxy_func,
    remote,
)

__all__ = [
    "BasePlugin",
    "PluginApi",
    "PluginMetadata",
    "get_plugin",
    "get_plugin_config",
    "load_builtin_plugins",
    "load_plugins",
    "logger",
    "on_proxy",
    "on_register",
    "on_request",
    "register_proxy_func",
    "remote",
]



"""Tests for framex.adapter.ray_adapter module."""

import sys
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from fastapi import FastAPI

from framex.adapter.base import AdapterMode


class TestRayAdapterImport:
    """Tests for RayAdapter import error handling."""

    def test_import_error_without_ray(self):
        """Test that importing RayAdapter without ray raises RuntimeError."""
        # Remove ray from sys.modules if it exists
        ray_modules = [key for key in sys.modules if key.startswith("ray")]
        saved_modules = {}
        for mod in ray_modules:
            saved_modules[mod] = sys.modules.pop(mod, None)

        try:
            with patch.dict("sys.modules", {"ray": None, "ray.serve": None}):  # noqa
                with pytest.raises(RuntimeError, match="Ray engine requires extra dependency"):  # noqa
                    # Force reload to trigger ImportError
                    import importlib

                    import framex.adapter.ray_adapter

                    importlib.reload(framex.adapter.ray_adapter)
        finally:
            # Restore modules
            for mod, val in saved_modules.items():
                if val is not None:
                    sys.modules[mod] = val


@pytest.fixture
def mock_ray():
    """Mock ray and ray.serve modules."""
    mock_ray_module = MagicMock()
    mock_serve_module = MagicMock()
    mock_deployment_handle = MagicMock()

    with (
        patch.dict(
            "sys.modules",
            {
                "ray": mock_ray_module,
                "ray.serve": mock_serve_module,
                "ray.serve.handle": MagicMock(DeploymentHandle=mock_deployment_handle),
            },
        ),
        patch("framex.adapter.ray_adapter.ray", mock_ray_module),
        patch("framex.adapter.ray_adapter.serve", mock_serve_module),
        patch("framex.adapter.ray_adapter.DeploymentHandle", mock_deployment_handle),
    ):
        yield mock_ray_module, mock_serve_module, mock_deployment_handle


class TestRayAdapter:
    """Tests for RayAdapter class with mocked ray dependencies."""

    def test_mode_is_ray(self, mock_ray):  # noqa
        """Test RayAdapter mode is RAY."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()
        assert adapter.mode == AdapterMode.RAY

    def test_to_ingress_calls_serve_ingress(self, mock_ray):
        """Test to_ingress calls serve.ingress and to_deployment."""
        from framex.adapter.ray_adapter import RayAdapter

        _, mock_serve_module, _ = mock_ray
        adapter = RayAdapter()

        class TestClass:
            pass

        app = FastAPI()

        # Setup mock chain
        mock_ingress_decorator = MagicMock()
        mock_ingress_result = MagicMock()
        mock_ingress_decorator.return_value = mock_ingress_result
        mock_serve_module.ingress.return_value = mock_ingress_decorator

        mock_deployment_decorator = MagicMock()
        mock_deployment_result = MagicMock()
        mock_deployment_decorator.return_value = mock_deployment_result
        mock_serve_module.deployment.return_value = mock_deployment_decorator

        adapter.to_ingress(TestClass, app, name="test")

        mock_serve_module.ingress.assert_called_once_with(app)
        mock_ingress_decorator.assert_called_once_with(TestClass)
        mock_serve_module.deployment.assert_called_once_with(name="test")

    def test_to_deployment_calls_serve_deployment(self, mock_ray):
        """Test to_deployment calls serve.deployment with kwargs."""
        from framex.adapter.ray_adapter import RayAdapter

        _, mock_serve_module, _ = mock_ray
        adapter = RayAdapter()

        class TestClass:
            pass

        mock_decorator = MagicMock()
        mock_result = MagicMock()
        mock_decorator.return_value = mock_result
        mock_serve_module.deployment.return_value = mock_decorator

        adapter.to_deployment(TestClass, name="test", num_replicas=3)

        mock_serve_module.deployment.assert_called_once_with(name="test", num_replicas=3)
        mock_decorator.assert_called_once_with(TestClass)

    def test_to_remote_func_with_sync_function(self, mock_ray):
        """Test to_remote_func wraps sync function with ray.remote."""
        from framex.adapter.ray_adapter import RayAdapter

        mock_ray_module, _, _ = mock_ray
        adapter = RayAdapter()

        def sync_func(x):
            return x * 2

        mock_remote_func = MagicMock()
        mock_ray_module.remote.return_value = mock_remote_func

        result = adapter.to_remote_func(sync_func)

        mock_ray_module.remote.assert_called_once()
        assert result == mock_remote_func

    def test_to_remote_func_with_async_function(self, mock_ray):
        """Test to_remote_func wraps async function to sync then ray.remote."""
        from framex.adapter.ray_adapter import RayAdapter

        mock_ray_module, _, _ = mock_ray
        adapter = RayAdapter()

        async def async_func(x):
            return x * 2

        mock_remote_func = MagicMock()
        mock_ray_module.remote.return_value = mock_remote_func

        result = adapter.to_remote_func(async_func)

        # Should wrap async function and call ray.remote
        mock_ray_module.remote.assert_called_once()
        assert result == mock_remote_func

    def test_to_remote_func_async_wrapper_runs_asyncio(self, mock_ray):
        """Test that async wrapper uses asyncio.run."""
        from framex.adapter.ray_adapter import RayAdapter

        mock_ray_module, _, _ = mock_ray
        adapter = RayAdapter()

        call_log = []

        async def async_func(x):
            call_log.append(f"called with {x}")
            return x * 2

        # Capture the wrapper function passed to ray.remote
        captured_wrapper = None

        def capture_remote(func):
            nonlocal captured_wrapper
            captured_wrapper = func
            return MagicMock()

        mock_ray_module.remote = capture_remote

        adapter.to_remote_func(async_func)

        # Now test the captured wrapper
        assert captured_wrapper is not None
        with patch("asyncio.run") as mock_asyncio_run:
            mock_asyncio_run.return_value = 10
            result = captured_wrapper(5)
            mock_asyncio_run.assert_called_once()
            assert result == 10

    def test_get_handle_calls_serve_get_deployment_handle(self, mock_ray):
        """Test get_handle calls serve.get_deployment_handle."""
        from framex.adapter.ray_adapter import RayAdapter

        _, mock_serve_module, _ = mock_ray
        adapter = RayAdapter()

        mock_handle = MagicMock()
        mock_serve_module.get_deployment_handle.return_value = mock_handle

        with patch("framex.adapter.ray_adapter.APP_NAME", "test_app"):
            result = adapter.get_handle("test_deployment")

        mock_serve_module.get_deployment_handle.assert_called_once_with("test_deployment", app_name="test_app")
        assert result == mock_handle

    def test_bind_calls_deployment_bind(self, mock_ray):  # noqa
        """Test bind calls deployment.bind with kwargs."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()

        mock_deployment = MagicMock()
        mock_bound = MagicMock()
        mock_deployment.bind.return_value = mock_bound

        result = adapter.bind(mock_deployment, param1="value1", param2="value2")

        mock_deployment.bind.assert_called_once_with(param1="value1", param2="value2")
        assert result == mock_bound

    def test_stream_call_uses_options_stream(self, mock_ray):  # noqa
        """Test _stream_call uses options(stream=True).remote."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()

        mock_func = MagicMock()
        mock_options = MagicMock()
        mock_remote_result = MagicMock()
        mock_func.options.return_value = mock_options
        mock_options.remote.return_value = mock_remote_result

        result = adapter._stream_call(mock_func, param="value")

        mock_func.options.assert_called_once_with(stream=True)
        mock_options.remote.assert_called_once_with(param="value")
        assert result == mock_remote_result

    async def test_acall_awaits_remote(self, mock_ray):
        """Test _acall awaits func.remote."""
        from framex.adapter.ray_adapter import RayAdapter

        _, _, _ = mock_ray
        adapter = RayAdapter()

        mock_func = MagicMock()
        mock_remote = AsyncMock(return_value="result")
        mock_func.remote = mock_remote

        result = await adapter._acall(mock_func, param="value")

        mock_remote.assert_called_once_with(param="value")
        assert result == "result"

    async def test_invoke_with_async_function(self, mock_ray):
        """Test _invoke delegates to _acall for async functions."""
        from framex.adapter.ray_adapter import RayAdapter

        _, _, mock_deployment_handle = mock_ray
        adapter = RayAdapter()

        async def async_func(**kwargs):
            return "async_result"

        # Patch isinstance to always return False for DeploymentHandle check
        with (
            patch(
                "framex.adapter.ray_adapter.isinstance",
                side_effect=lambda obj, cls: False if cls is mock_deployment_handle else isinstance(obj, cls),
            ),
            patch.object(adapter, "_acall", new=AsyncMock(return_value="async_result")) as mock_acall,
        ):
            result = await adapter._invoke(async_func, param="value")
            mock_acall.assert_called_once_with(async_func, param="value")
            assert result == "async_result"

    async def test_invoke_with_deployment_handle(self, mock_ray):
        """Test _invoke delegates to _acall for DeploymentHandle-like object."""
        from framex.adapter.ray_adapter import RayAdapter

        _, _, mock_deployment_handle = mock_ray
        adapter = RayAdapter()

        # Create a mock that will match isinstance check
        # We'll patch inspect.iscoroutinefunction to return False
        # and the isinstance(func, DeploymentHandle) will be checked
        mock_handle = MagicMock()

        with (
            patch("inspect.iscoroutinefunction", return_value=False),
            patch(
                "framex.adapter.ray_adapter.isinstance",
                side_effect=lambda obj, cls: obj is mock_handle and cls is mock_deployment_handle,
            ),
            patch.object(adapter, "_acall", new=AsyncMock(return_value="handle_result")) as mock_acall,
        ):
            result = await adapter._invoke(mock_handle, param="value")
            mock_acall.assert_called_once_with(mock_handle, param="value")
            assert result == "handle_result"

    async def test_invoke_with_sync_function(self, mock_ray):
        """Test _invoke delegates to _call for sync functions."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()
        _, _, mock_deployment_handle = mock_ray

        def sync_func(**kwargs):
            return "sync_result"

        # Patch isinstance to always return False for DeploymentHandle check
        with (
            patch(
                "framex.adapter.ray_adapter.isinstance",
                side_effect=lambda obj, cls: False if cls is mock_deployment_handle else isinstance(obj, cls),
            ),
            patch.object(adapter, "_call", return_value="sync_result") as mock_call,
        ):
            result = await adapter._invoke(sync_func, param="value")
            mock_call.assert_called_once_with(sync_func, param="value")
            assert result == "sync_result"

    def test_call_invokes_function_directly(self, mock_ray):  # noqa
        """Test _call invokes function directly with kwargs."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()

        def mock_func(**kwargs):
            return kwargs

        result = adapter._call(mock_func, key1="value1", key2="value2")
        assert result == {"key1": "value1", "key2": "value2"}

    def test_bind_with_no_kwargs(self, mock_ray):  # noqa
        """Test bind with no kwargs."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()

        mock_deployment = MagicMock()
        mock_bound = MagicMock()
        mock_deployment.bind.return_value = mock_bound

        result = adapter.bind(mock_deployment)

        mock_deployment.bind.assert_called_once_with()
        assert result == mock_bound

    def test_stream_call_with_multiple_kwargs(self, mock_ray):  # noqa
        """Test _stream_call passes all kwargs correctly."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()

        mock_func = MagicMock()
        mock_options = MagicMock()
        mock_remote_result = MagicMock()
        mock_func.options.return_value = mock_options
        mock_options.remote.return_value = mock_remote_result

        result = adapter._stream_call(mock_func, a=1, b=2, c=3)

        mock_options.remote.assert_called_once_with(a=1, b=2, c=3)
        assert result == mock_remote_result

    def test_to_deployment_with_no_kwargs(self, mock_ray):
        """Test to_deployment with no kwargs."""
        from framex.adapter.ray_adapter import RayAdapter

        _, mock_serve_module, _ = mock_ray
        adapter = RayAdapter()

        class TestClass:
            pass

        mock_decorator = MagicMock()
        mock_result = MagicMock()
        mock_decorator.return_value = mock_result
        mock_serve_module.deployment.return_value = mock_decorator

        adapter.to_deployment(TestClass)

        mock_serve_module.deployment.assert_called_once_with()
        mock_decorator.assert_called_once_with(TestClass)

    async def test_acall_with_no_kwargs(self, mock_ray):  # noqa
        """Test _acall with no kwargs."""
        from framex.adapter.ray_adapter import RayAdapter

        adapter = RayAdapter()

        mock_func = MagicMock()
        mock_remote = AsyncMock(return_value="result")
        mock_func.remote = mock_remote

        result = await adapter._acall(mock_func)

        mock_remote.assert_called_once_with()
        assert result == "result"

    def test_get_handle_with_different_app_name(self, mock_ray):
        """Test get_handle uses APP_NAME constant correctly."""
        from framex.adapter.ray_adapter import RayAdapter

        _, mock_serve_module, _ = mock_ray
        adapter = RayAdapter()

        mock_handle = MagicMock()
        mock_serve_module.get_deployment_handle.return_value = mock_handle

        with patch("framex.adapter.ray_adapter.APP_NAME", "custom_app"):
            result = adapter.get_handle("my_deployment")

        mock_serve_module.get_deployment_handle.assert_called_once_with("my_deployment", app_name="custom_app")
        assert result == mock_handle



"""Comprehensive tests for framex.driver.application module."""

from typing import Any
from unittest.mock import patch

import pytest
from fastapi.testclient import TestClient
from starlette import status
from starlette.exceptions import HTTPException

from framex.consts import API_STR
from framex.driver.application import create_fastapi_application


class TestCreateFastAPIApplication:
    """Test suite for create_fastapi_application function."""

    def test_openapi_url_is_none_by_default(self):
        app = create_fastapi_application()
        assert app.openapi_url is None

    def test_docs_url_is_none_by_default(self):
        app = create_fastapi_application()
        assert app.docs_url is None

    def test_redirect_slashes_disabled(self):
        app = create_fastapi_application()
        assert app.router.redirect_slashes is False

    def test_cors_middleware_configured(self):
        app = create_fastapi_application()
        middleware_classes = [m.cls.__name__ for m in app.user_middleware]  # type: ignore
        assert "CORSMiddleware" in middleware_classes


class TestExceptionHandlers:
    @pytest.fixture
    def app(self):
        return create_fastapi_application()

    @pytest.fixture
    def client(self, app):
        return TestClient(app, raise_server_exceptions=False)

    def test_http_exception_handler(self, app, client):
        @app.get("/test-http-exception")
        async def endpoint() -> None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail="Not found",
            )

        response = client.get("/test-http-exception")
        data = response.json()

        assert response.status_code == 404
        assert data["status"] == 404
        assert data["message"] == "Not found"
        assert data["is_middleware_error"] is True

    def test_general_exception_handler_returns_500(self, app, client):
        @app.get("/test-general-exception")
        async def endpoint() -> None:
            raise ValueError("Something went wrong")

        response = client.get("/test-general-exception")
        data = response.json()

        assert response.status_code == 500
        assert data["status"] == 500
        assert "Something went wrong" in data["message"]
        assert "timestamp" in data

    def test_unicode_exception_message(self, app, client):
        @app.get("/test-unicode")
        async def endpoint() -> None:
            raise RuntimeError("错误信息 你好 🌍")

        response = client.get("/test-unicode")
        data = response.json()

        assert response.status_code == 500
        assert "你好" in data["message"]
        assert "🌍" in data["message"]


class TestLogResponseMiddleware:
    @pytest.fixture
    def app(self):
        return create_fastapi_application()

    @pytest.fixture
    def client(self, app):
        return TestClient(app, raise_server_exceptions=False)

    def test_api_response_wrapped(self, app, client):
        @app.get(f"{API_STR}/test-wrap")
        async def endpoint() -> Any:
            return {"result": "ok"}

        response = client.get(f"{API_STR}/test-wrap")
        data = response.json()

        assert response.status_code == 200
        assert data["status"] == 200
        assert data["message"] == "success"
        assert data["data"] == {"result": "ok"}
        assert "timestamp" in data


class TestLifespanBehavior:
    @patch("framex.config.settings")
    def test_lifespan_without_ray(self, mock_settings):
        mock_settings.server.use_ray = False
        app = create_fastapi_application()
        assert app.router.lifespan_context is not None

    @patch("framex.config.settings")
    def test_lifespan_with_ray(self, mock_settings):
        mock_settings.server.use_ray = True
        app = create_fastapi_application()
        assert app.router.lifespan_context is not None



import json
from datetime import datetime, timedelta
from typing import Any

import pytest
from pydantic import BaseModel

from framex.config import AuthConfig
from framex.utils import (
    StreamEnventType,
    cache_decode,
    cache_encode,
    format_uptime,
    make_stream_event,
    safe_error_message,
)


class StreamDataModel(BaseModel):
    content: str
    id: int


@pytest.mark.parametrize(
    ("event_type", "data", "result"),
    [
        ("event a", "data a", 'event: event a\ndata: {"content": "data a"}\n\n'),
        (
            StreamEnventType.MESSAGE_CHUNK,
            {"result": "chunk data"},
            'event: message_chunk\ndata: {"result": "chunk data"}\n\n',
        ),
        (
            StreamEnventType.DEBUG,
            StreamDataModel(content="data a", id=1),
            'event: debug\ndata: {"content": "data a", "id": 1}\n\n',
        ),
    ],
)
def test_make_stream_event(event_type: StreamEnventType | str, data: str | dict[str, Any] | BaseModel, result: str):
    res = make_stream_event(event_type, data)
    assert res == result


def test_is_url_protected():
    cfg = AuthConfig(
        rules={
            "/api/user": ["k1"],
            "/api/*": ["k2"],
        }
    )

    assert cfg._is_url_protected("/api/user")
    assert cfg._is_url_protected("/api/user/1")
    assert not cfg._is_url_protected("/admin")


def test_get_auth_keys():
    cfg = AuthConfig(
        rules={
            "/api/*": ["k1"],
            "/api/admin/*": ["k2"],
            "/api/admin/user": ["k3"],
        }
    )

    assert cfg.get_auth_keys("/public") is None
    assert cfg.get_auth_keys("/api/user") == ["k1"]
    assert cfg.get_auth_keys("/api/admin/test") == ["k2"]
    assert cfg.get_auth_keys("/api/admin/user") == ["k3"]


class SubModel(BaseModel):
    id: int
    name: str


class ExchangeModel(BaseModel):
    id: str
    name: int
    model: SubModel
    created_at: datetime


def test_basic_types():
    data = {"a": 1, "b": "string", "c": [1, 2, 3], "d": {"nested": True}}
    encoded = cache_encode(data)
    decoded = cache_decode(encoded)
    assert decoded == data
    assert isinstance(decoded["c"], list)


def test_datetime_and_enum():
    now = datetime(2025, 12, 23, 10, 0, 0)
    data = {"time": now}
    encoded = cache_encode(data)
    decoded = cache_decode(encoded)
    assert "2025-12-23T10:00:00" in str(decoded["time"])


def test_nested_pydantic_models():
    sub = SubModel(id=1, name="sub_name")
    main = ExchangeModel(id="main_id", name=100, model=sub, created_at=datetime.now())

    original_data = {"status": "success", "result_list": [main, main], "single_model": main}

    encoded = cache_encode(original_data)
    assert isinstance(encoded, str)

    decoded = cache_decode(encoded)

    res_model = decoded["single_model"]
    assert res_model.id == "main_id"
    assert res_model.name == 100

    assert isinstance(res_model.model, SubModel)
    assert res_model.model.id == 1
    assert res_model.model.name == "sub_name"

    assert decoded["result_list"][0].model.id == 1


def test_recovery_failure_fallback():
    fake_payload = {
        "__type__": "dynamic_obj",
        "__module__": "non.existent.module",
        "__class__": "MissingClass",
        "data": {"id": 999, "info": "test"},
    }
    encoded = json.dumps(fake_payload)
    decoded = cache_decode(encoded)

    assert decoded.id == 999
    assert decoded.info == "test"


def test_format_uptime():
    """Test format_uptime function"""
    # Test seconds only
    delta = timedelta(seconds=45)
    assert format_uptime(delta) == "45s"

    # Test minutes and seconds
    delta = timedelta(seconds=125)
    assert format_uptime(delta) == "2m 5s"

    # Test hours, minutes, and seconds
    delta = timedelta(seconds=3665)
    assert format_uptime(delta) == "1h 1m 5s"

    # Test days
    delta = timedelta(days=2, seconds=3661)
    assert format_uptime(delta) == "2d 1h 1m 1s"

    # Test zero seconds
    delta = timedelta(seconds=0)
    assert format_uptime(delta) == "0s"

    # Test only minutes (no seconds)
    delta = timedelta(minutes=5, seconds=0)
    assert format_uptime(delta) == "5m"


class CauseError(Exception):
    def __init__(self, cause: Exception):
        self.cause = cause
        super().__init__("outer")


def test_safe_error_message_with_cause():
    e = CauseError(RuntimeError("inner"))
    assert safe_error_message(e) == "inner"


def test_safe_error_message_with_args():
    e = RuntimeError("simple error")
    assert safe_error_message(e) == "simple error"


def test_safe_error_message_fallback():
    e = Exception()
    e.args = ()
    assert safe_error_message(e) == "Internal Server Error"



import base64
import importlib
import inspect
import json
import re
import zlib
from collections.abc import Callable
from datetime import datetime, timedelta
from enum import Enum, StrEnum
from itertools import cycle
from pathlib import Path
from typing import Any

from pydantic import BaseModel


def plugin_to_deployment_name(plugin_name: str, obj_name: str) -> str:
    return f"{plugin_name}.{obj_name}"


def path_to_module_name(path: Path) -> str:
    """Convert path to module name"""
    rel_path = path.resolve().relative_to(Path.cwd().resolve())
    if rel_path.stem == "__init__":
        module_name = ".".join(rel_path.parts[:-1])
    else:
        module_name = ".".join([*rel_path.parts[:-1], rel_path.stem])  # type: ignore
    return module_name.removeprefix("src.")


def escape_tag(s: str) -> str:
    """Used to escape `<tag>` type special tags when recording color logs"""
    return re.sub(r"</?((?:[fb]g\s)?[^<>\s]*)>", r"\\\g<0>", s)


def extract_method_params(func: Callable) -> list[tuple[str, Any]]:
    sig = inspect.signature(func)
    params = []
    for param in sig.parameters.values():
        if param.name == "self":
            continue
        params.append((param.name, param.annotation))
    return params


class StreamEnventType(StrEnum):
    MESSAGE_CHUNK = "message_chunk"
    FINISH = "finish"
    ERROR = "error"
    DEBUG = "debug"


def make_stream_event(event_type: StreamEnventType | str, data: str | dict[str, Any] | BaseModel | None = None) -> str:
    if not data:
        data = {}
    elif isinstance(data, BaseModel):
        data = data.model_dump()
    elif isinstance(data, str):
        data = {"content": data}
    return f"event: {event_type}\ndata: {json.dumps(data, ensure_ascii=False)}\n\n"


def xor_crypt(data: bytes, key: str = "01234567890abcdefghijklmnopqrstuvwxyz") -> bytes:
    return bytes(a ^ b for a, b in zip(data, cycle(key.encode())))


def cache_encode(data: Any) -> str:
    def transform(obj: Any) -> Any:
        if hasattr(obj, "__dict__"):
            raw_attributes = {k: transform(v) for k, v in obj.__dict__.items() if not k.startswith("_")}
            return {
                "__type__": "dynamic_obj",
                "__module__": obj.__class__.__module__,
                "__class__": obj.__class__.__name__,
                "data": raw_attributes,
            }
        if isinstance(obj, list):
            return [transform(i) for i in obj]
        if isinstance(obj, dict):
            return {k: transform(v) for k, v in obj.items()}
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, Enum):
            return obj.value
        return obj

    json_str = json.dumps(transform(data), ensure_ascii=False)
    compressed = zlib.compress(json_str.encode("utf-8"))
    encrypted = xor_crypt(compressed)
    return base64.b64encode(encrypted).decode("ascii")


def cache_decode(res: Any) -> Any:
    current = res
    while isinstance(current, str):
        try:
            decoded_bytes = base64.b64decode(current, validate=True)
            current = zlib.decompress(xor_crypt(decoded_bytes)).decode("utf-8")
        except Exception:
            try:
                temp = json.loads(current)
                if temp == current:
                    break
                current = temp
            except Exception:
                break

    def restore_models(item: Any) -> Any:
        if isinstance(item, list):
            return [restore_models(i) for i in item]

        if isinstance(item, dict):
            if item.get("__type__") == "dynamic_obj":
                try:
                    module = importlib.import_module(item["__module__"])
                    cls = getattr(module, item["__class__"])

                    cleaned_data = {k: restore_models(v) for k, v in item["data"].items()}

                    if hasattr(cls, "model_validate"):
                        return cls.model_validate(cleaned_data)
                    return cls(**cleaned_data)
                except Exception:
                    from types import SimpleNamespace

                    return SimpleNamespace(**{k: restore_models(v) for k, v in item["data"].items()})

            return {k: restore_models(v) for k, v in item.items()}

        return item

    return restore_models(current)


def format_uptime(delta: timedelta) -> str:
    days = delta.days
    hours, remainder = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    parts = []
    if days:
        parts.append(f"{days}d")
    if hours:
        parts.append(f"{hours}h")
    if minutes:
        parts.append(f"{minutes}m")
    if seconds or not parts:
        parts.append(f"{seconds}s")

    return " ".join(parts)


def safe_error_message(e: Exception) -> str:
    if hasattr(e, "cause") and e.cause:
        return str(e.cause)
    if e.args:
        return str(e.args[0])
    return "Internal Server Error"


def shorten_str(data: str, max_len: int = 45) -> str:
    return data if len(data) <= max_len else data[: max_len - 3] + "..."



from framex.adapter.base import BaseAdapter
from framex.adapter.local_adapter import LocalAdapter
from framex.config import settings

_adapter: BaseAdapter | None = None


def get_adapter() -> BaseAdapter:
    global _adapter
    if _adapter is None:
        if settings.server.use_ray:
            from framex.adapter.ray_adapter import RayAdapter

            _adapter = RayAdapter()
        else:
            _adapter = LocalAdapter()
    return _adapter



from typing import Any, Self

from pydantic import BaseModel, Field, model_validator

from framex.config import AuthConfig
from framex.plugin import get_plugin_config


class ProxyPluginConfig(BaseModel):
    proxy_urls: list[str] = Field(default_factory=list)
    force_stream_apis: list[str] = Field(default_factory=list)
    timeout: int = 600
    ingress_config: dict[str, Any] = Field(default_factory=lambda: {"max_ongoing_requests": 60})

    white_list: list[str] = Field(default_factory=list)

    auth: AuthConfig = Field(default_factory=AuthConfig)

    proxy_functions: dict[str, list[str]] = Field(default_factory=dict)

    def is_white_url(self, url: str) -> bool:
        """Check if a URL is protected by any auth_urls rule."""
        if self.white_list == []:  # pragma: no cover
            return True
        for rule in self.white_list:
            if rule == url:
                return True
            if rule.endswith("/*") and url.startswith(rule[:-1]):
                return True
        return False

    @model_validator(mode="after")
    def validate_proxy_functions(self) -> Self:
        for url in self.proxy_functions:
            if url not in self.proxy_urls:  # pragma: no cover
                raise ValueError(f"proxy_functions url '{url}' is not covered by any proxy_urls rule")
        return self


settings: ProxyPluginConfig = get_plugin_config("proxy", ProxyPluginConfig)



"""Tests for framex.adapter.local_adapter module."""

import asyncio
import threading
from unittest.mock import MagicMock, patch

from framex.adapter.base import AdapterMode
from framex.adapter.local_adapter import LocalAdapter


class TestLocalAdapter:
    """Tests for LocalAdapter class."""

    def test_mode_is_local(self):
        """Test LocalAdapter mode is LOCAL."""
        adapter = LocalAdapter()
        assert adapter.mode == AdapterMode.LOCAL

    def test_to_deployment_sets_deployment_name(self):
        """Test to_deployment sets deployment_name attribute on class."""
        adapter = LocalAdapter()

        class TestClass:
            pass

        result = adapter.to_deployment(TestClass, name="my_deployment")
        assert hasattr(result, "deployment_name")
        assert result.deployment_name == "my_deployment"

    def test_to_deployment_without_name(self):
        """Test to_deployment without name doesn't set attribute."""
        adapter = LocalAdapter()

        class TestClass:
            pass

        result = adapter.to_deployment(TestClass)
        assert not hasattr(result, "deployment_name")

    def test_to_deployment_with_other_kwargs(self):
        """Test to_deployment handles other kwargs."""
        adapter = LocalAdapter()

        class TestClass:
            pass

        result = adapter.to_deployment(TestClass, name="test", other_param="value")
        assert hasattr(result, "deployment_name")
        assert result.deployment_name == "test"

    def test_get_handle_returns_deployment(self):
        """Test get_handle returns deployment from app state."""
        adapter = LocalAdapter()

        mock_deployment = MagicMock()
        mock_app = MagicMock()
        mock_app.state.deployments_dict = {"test_deployment": mock_deployment}

        with patch("framex.driver.ingress.app", mock_app):
            handle = adapter.get_handle("test_deployment")
            assert handle == mock_deployment

    def test_get_handle_returns_ingress_for_backend(self):
        """Test get_handle returns ingress for BACKEND_NAME."""
        adapter = LocalAdapter()

        mock_ingress = MagicMock()
        mock_app = MagicMock()
        mock_app.state.deployments_dict = {}
        mock_app.state.ingress = mock_ingress

        with (
            patch("framex.driver.ingress.app", mock_app),
            patch("framex.consts.BACKEND_NAME", "backend"),
        ):
            handle = adapter.get_handle("backend")
            assert handle == mock_ingress

    def test_get_handle_returns_none_when_not_found(self):
        """Test get_handle returns None when deployment not found."""
        adapter = LocalAdapter()

        mock_app = MagicMock()
        mock_app.state.deployments_dict = {}

        with patch("framex.driver.ingress.app", mock_app):
            handle = adapter.get_handle("nonexistent")
            assert handle is None

    def test_bind_calls_deployment_with_kwargs(self):
        """Test bind calls deployment function with kwargs."""
        adapter = LocalAdapter()

        def mock_deployment(**kwargs):
            return kwargs

        result = adapter.bind(mock_deployment, param1="value1", param2="value2")
        assert result == {"param1": "value1", "param2": "value2"}

    def test_bind_with_no_kwargs(self):
        """Test bind calls deployment with no kwargs."""
        adapter = LocalAdapter()

        def mock_deployment(**kwargs):
            return "called"

        result = adapter.bind(mock_deployment)
        assert result == "called"

    def test_safe_plot_wrapper_uses_lock(self):
        """Test _safe_plot_wrapper uses thread lock."""
        adapter = LocalAdapter()
        call_order = []
        lock_acquired = []

        def mock_func(*args, **kwargs):
            # Check if lock is held
            lock_acquired.append(threading.current_thread().ident)
            call_order.append("func")
            return "result"

        result = adapter._safe_plot_wrapper(mock_func, "arg1", kwarg1="value1")
        assert result == "result"
        assert len(lock_acquired) == 1

    def test_safe_plot_wrapper_serializes_calls(self):
        """Test _safe_plot_wrapper serializes concurrent calls."""
        adapter = LocalAdapter()
        call_order = []
        lock = threading.Lock()  # noqa

        def slow_func(name):
            call_order.append(f"{name}_start")
            threading.Event().wait(0.01)  # Small delay
            call_order.append(f"{name}_end")
            return name

        # Run two calls concurrently
        def run_wrapper(name):
            return adapter._safe_plot_wrapper(slow_func, name)

        thread1 = threading.Thread(target=run_wrapper, args=("call1",))
        thread2 = threading.Thread(target=run_wrapper, args=("call2",))

        thread1.start()
        thread2.start()
        thread1.join()
        thread2.join()

        # Both calls should complete (in any order due to thread scheduling)
        assert "call1_start" in call_order
        assert "call1_end" in call_order
        assert "call2_start" in call_order
        assert "call2_end" in call_order

    async def test_to_remote_func_with_async_function(self):
        """Test to_remote_func handles async functions."""
        adapter = LocalAdapter()

        async def async_func(value):
            return value * 2

        wrapped_func = adapter.to_remote_func(async_func)
        assert hasattr(wrapped_func, "remote")

        result = await wrapped_func.remote(5)
        assert result == 10

    async def test_to_remote_func_with_sync_function(self):
        """Test to_remote_func handles sync functions with asyncio.to_thread."""
        adapter = LocalAdapter()

        def sync_func(value):
            return value * 3

        wrapped_func = adapter.to_remote_func(sync_func)
        assert hasattr(wrapped_func, "remote")

        result = await wrapped_func.remote(5)
        assert result == 15

    async def test_to_remote_func_with_sync_function_uses_safe_wrapper(self):
        """Test to_remote_func uses _safe_plot_wrapper for sync functions."""
        adapter = LocalAdapter()

        def sync_func(value):
            return value + 1

        with patch.object(adapter, "_safe_plot_wrapper", return_value=11) as mock_wrapper:
            wrapped_func = adapter.to_remote_func(sync_func)
            result = await wrapped_func.remote(10)  # type: ignore

            # Verify safe wrapper was used
            mock_wrapper.assert_called_once_with(sync_func, 10)
            assert result == 11

    async def test_invoke_with_async_function(self):
        """Test _invoke delegates to _acall for async functions."""
        adapter = LocalAdapter()

        async def async_func(**kwargs):
            return "async_result"

        result = await adapter._invoke(async_func, param="value")
        assert result == "async_result"

    async def test_invoke_with_sync_function(self):
        """Test _invoke delegates to _call for sync functions."""
        adapter = LocalAdapter()

        def sync_func(**kwargs):
            return "sync_result"

        result = await adapter._invoke(sync_func, param="value")
        assert result == "sync_result"

    async def test_acall_awaits_async_function(self):
        """Test _acall awaits async function with kwargs."""
        adapter = LocalAdapter()

        async def async_func(**kwargs):
            return kwargs

        result = await adapter._acall(async_func, key1="value1", key2="value2")
        assert result == {"key1": "value1", "key2": "value2"}

    def test_call_invokes_sync_function(self):
        """Test _call invokes sync function with kwargs."""
        adapter = LocalAdapter()

        def sync_func(**kwargs):
            return kwargs

        result = adapter._call(sync_func, key1="value1", key2="value2")
        assert result == {"key1": "value1", "key2": "value2"}

    def test_stream_call_invokes_function(self):
        """Test _stream_call invokes function with kwargs."""
        adapter = LocalAdapter()

        def stream_func(**kwargs):
            yield from kwargs.values()

        result = adapter._stream_call(stream_func, key1="value1", key2="value2")
        values = list(result)
        assert "value1" in values
        assert "value2" in values

    async def test_stream_call_with_async_generator(self):
        """Test _stream_call works with async generators."""
        adapter = LocalAdapter()

        async def async_stream(**kwargs):
            for value in kwargs.values():
                yield value

        result = adapter._stream_call(async_stream, key1="value1", key2="value2")
        values = []
        async for value in result:
            values.append(value)  # noqa
        assert "value1" in values
        assert "value2" in values

    def test_to_remote_func_preserves_original_function(self):
        """Test to_remote_func preserves the original function."""
        adapter = LocalAdapter()

        def original_func(x):
            return x * 2

        wrapped_func = adapter.to_remote_func(original_func)
        # Original function should still be callable
        assert wrapped_func(5) == 10

    async def test_to_remote_func_remote_attribute(self):
        """Test to_remote_func adds remote attribute."""
        adapter = LocalAdapter()

        def sync_func(x):
            return x + 1

        wrapped_func = adapter.to_remote_func(sync_func)
        assert hasattr(wrapped_func, "remote")
        assert callable(wrapped_func.remote)

    async def test_concurrent_safe_plot_wrapper_calls(self):
        """Test multiple concurrent calls to _safe_plot_wrapper are serialized."""
        adapter = LocalAdapter()
        results = []

        def counting_func(n):
            import time

            time.sleep(0.001)  # Small delay to ensure overlap without lock
            results.append(n)
            return n

        async def async_wrapper(n):
            return await asyncio.to_thread(adapter._safe_plot_wrapper, counting_func, n)

        # Run multiple concurrent calls
        await asyncio.gather(*[async_wrapper(i) for i in range(5)])

        # All calls should complete
        assert len(results) == 5
        assert set(results) == {0, 1, 2, 3, 4}

    def test_get_handle_with_empty_deployments_dict(self):
        """Test get_handle with empty deployments dict."""
        adapter = LocalAdapter()

        mock_app = MagicMock()
        mock_app.state.deployments_dict = {}
        mock_app.state.ingress = None

        with patch("framex.driver.ingress.app", mock_app):
            handle = adapter.get_handle("any_deployment")
            assert handle is None

    async def test_invoke_preserves_kwargs(self):
        """Test _invoke preserves kwargs correctly."""
        adapter = LocalAdapter()

        async def async_func(**kwargs):
            return kwargs

        result = await adapter._invoke(async_func, a=1, b=2, c=3)
        assert result == {"a": 1, "b": 2, "c": 3}



import re
from collections.abc import Callable
from enum import Enum
from typing import Any

from fastapi import Depends, HTTPException, Request, Response, status
from fastapi.responses import JSONResponse, StreamingResponse
from fastapi.routing import APIRoute
from pydantic import create_model
from starlette.routing import Route

from framex.adapter import get_adapter
from framex.consts import BACKEND_NAME
from framex.driver.application import create_fastapi_application
from framex.driver.auth import api_key_header, auth_jwt
from framex.driver.decorator import api_ingress
from framex.log import setup_logger
from framex.plugin.model import ApiType, PluginApi
from framex.utils import escape_tag, shorten_str

app = create_fastapi_application()


@app.get("/health")
async def health() -> str:
    return "ok"


@api_ingress(app=app, name=BACKEND_NAME)
class APIIngress:
    def __init__(self, deployments: list[Any], plugin_apis: list["PluginApi"]) -> None:
        setup_logger()
        app.state.ingress = self
        self.deployments_dict = {dep.deployment_name: dep for dep in deployments}
        app.state.deployments_dict = self.deployments_dict
        for plugin_api in plugin_apis:
            if (
                plugin_api.api
                and (deployment := self.deployments_dict.get(plugin_api.deployment_name))
                and plugin_api.call_type
                in [
                    ApiType.HTTP,
                    ApiType.ALL,
                ]
            ):
                self.register_route(
                    plugin_api.api,
                    plugin_api.methods,
                    plugin_api.func_name,
                    plugin_api.params,
                    deployment,
                    stream=plugin_api.stream,
                    direct_output=False,
                    tags=plugin_api.tags,
                )

    def register_route(
        self,
        path: str,
        methods: list[str],
        func_name: str,
        params: list[tuple[str, type | Callable]],
        handle: Any,
        stream: bool = False,
        direct_output: bool = False,
        tags: list[str | Enum] | None = None,
        auth_keys: list[str] | None = None,
    ) -> bool:
        from framex.log import logger

        if tags is None:
            tags = ["default"]
        if auth_keys is None:
            from framex.config import settings

            auth_keys = settings.auth.get_auth_keys(path)
            logger.trace(f"API({path}) with tags {tags} requires auth_keys {auth_keys}")
        adapter = get_adapter()

        try:
            routes: list[str] = [route.path for route in app.routes if isinstance(route, Route | APIRoute)]
            # logger.warning(f"API({path}) with tags {tags} is already registered, skipping duplicate registration.")
            methods_str = ",".join(m.upper() for m in methods)

            if path in routes:
                logger.opt(colors=True).warning(
                    f"API route already registered: {methods_str:<4} {path[:40] + '...':<45} ({handle.deployment_name})"
                )
                return False
            if (not path) or (not methods):
                raise RuntimeError(f"Api({path}) or methods({methods}) is empty")
            Model: BaseModel = create_model(f"{func_name}_InputModel", **{name: (tp, ...) for name, tp in params})  # type:ignore # noqa

            async def route_handler(response: Response, model: Model = Depends()) -> Any:  # type: ignore [valid-type]
                c_handle = getattr(handle, func_name)
                if not c_handle:
                    raise RuntimeError(
                        f"No handle found for api({methods}): {path} from {handle.deployment_name}.{func_name}"
                    )
                response.headers["X-Raw-Output"] = str(direct_output)
                if stream:
                    gen = adapter._stream_call(c_handle, **(model.__dict__))
                    return StreamingResponse(  # type: ignore
                        gen,
                        media_type="text/event-stream",
                    )

                return await adapter._acall(c_handle, **model.__dict__)  # type: ignore

            # Inject auth dependency if needed
            dependencies = []
            if auth_keys is not None:
                logger.trace(f"API({path}) with tags {tags} requires auth.")

                def _verify_api_key(request: Request, api_key: str | None = Depends(api_key_header)) -> None:
                    if (api_key is None or api_key not in auth_keys) and (not auth_jwt(request)):
                        logger.opt(colors=True).error(
                            f"<r>Unauthorized access attempt with API Key({api_key}) for API({path})</r>"
                        )
                        raise HTTPException(
                            status_code=status.HTTP_401_UNAUTHORIZED,
                            detail=f"Invalid API Key({api_key}) for API({path})",
                        )

                dependencies.append(Depends(_verify_api_key))

            self.add_api_route(
                path,
                route_handler,
                methods=methods,
                tags=tags,
                response_class=StreamingResponse if stream else JSONResponse,
                dependencies=dependencies,
            )
            methods_str = ",".join(m.upper() for m in methods)
            short_path = shorten_str(path)
            logger.opt(colors=True).success(
                f"API route registered: {methods_str:<4} <g>{short_path:<45}</g> ({handle.deployment_name})"
            )
            return True
        except Exception as e:
            logger.opt(exception=e, colors=True).error(
                f'<r>Failed to register api "{escape_tag(path)}" from {handle.deployment_name}</r>'
            )
        return False

    @app.get("/ping")
    async def inner(self) -> str:  # pragma: no cover
        return "pong"

    def __repr__(self):
        return BACKEND_NAME

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        methods: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        method_set: set[str] = {m.upper() for m in methods} if methods else {"GET"}
        norm_path = re.sub(r"\{[^}]+\}", "{}", path)

        for route in app.routes:
            if (
                isinstance(route, APIRoute)
                and re.sub(r"\{[^}]+\}", "{}", route.path) == norm_path
                and route.methods & method_set
            ):
                raise RuntimeError(f"Duplicate API route: {sorted(method_set)} {norm_path}")

        app.add_api_route(path, endpoint, methods=list(method_set), **kwargs)



from typing import Any
from unittest.mock import MagicMock

from framex.utils import cache_decode
from tests.consts import MOCK_RESPONSE


async def mock_get(_, url: str, *__, **kwargs: Any):
    resp = MagicMock()
    resp.raise_for_status.return_value = None

    headers = kwargs.get("headers", {})
    if headers.get("Authorization") != "i_am_proxy_docs_auth_keys":
        resp.json.return_value = {
            "status": 401,
            "message": f"Invalid API Key({headers.get('Authorization')}) for API(/api/v1/proxy/mock/auth/get)",
        }
    elif url.endswith("/api/v1/openapi.json"):
        resp.json.return_value = MOCK_RESPONSE
    else:
        raise AssertionError(f"Unexpected request: {url}")

    return resp


async def mock_request(_, method: str, url: str, **kwargs: Any):
    params = kwargs.get("params")
    body = kwargs.get("json") or kwargs.get("data")
    headers = kwargs.get("headers", {})

    resp = MagicMock()
    resp.raise_for_status.return_value = None
    if url.endswith("/proxy/mock/get") and method == "GET":
        resp.json.return_value = {
            "method": "GET",
            "params": params,
        }
    elif url.endswith("/proxy/mock/post") and method == "POST":
        resp.json.return_value = {
            "method": "POST",
            "params": params,
        }
    elif url.endswith("/proxy/mock/post_model") and method == "POST":
        resp.json.return_value = {
            "method": "POST",
            "body": body,
        }
    elif url.endswith("/proxy/mock/info") and method == "GET":
        resp.json.return_value = {
            "info": "i_am_mock_proxy_info",
        }
    elif url.endswith("/proxy/mock/auth/get") and method == "GET":
        if headers.get("Authorization") != "i_am_proxy_general_auth_keys":
            resp.json.return_value = {
                "status": 401,
                "message": f"Invalid API Key({headers.get('Authorization')}) for API(/api/v1/proxy/mock/auth/get)",
            }
        else:
            resp.json.return_value = {
                "method": "GET",
                "params": params,
            }
    elif url.endswith("/proxy/mock/auth/sget") and method == "GET":
        if headers.get("Authorization") != "i_am_proxy_special_auth_keys":
            resp.json.return_value = {
                "status": 401,
                "message": f"Invalid API Key({headers.get('Authorization')}) for API(/api/v1/proxy/mock/auth/get)",
            }
        else:
            resp.json.return_value = {
                "method": "GET",
                "params": params,
            }
    elif url.endswith("/proxy/remote") and method == "POST":
        if headers.get("Authorization") != "i_am_proxy_func_auth_keys":
            resp.json.return_value = {
                "status": 401,
                "message": f"Invalid API Key({headers.get('Authorization')}) for API(/api/v1/proxy/mock/auth/get)",
            }
        else:
            json_data = kwargs.get("json", {})
            func_name = json_data.get("func_name")
            data = json_data.get("data", {})
            if not func_name or not data:
                raise ValueError("Missing required fields: func_name and data")
            decode_func_name = cache_decode(func_name)
            decode_data = cache_decode(data)
            res = {"result": decode_func_name, "data": decode_data}
            resp.json.return_value = {"status": 200, "data": res}
    else:
        raise AssertionError(f"Unexpected request: {method} {url}")

    return resp



from collections.abc import Callable

from framex.consts import PROXY_PLUGIN_NAME
from framex.log import logger
from framex.plugin.model import ApiType, Plugin, PluginApi
from framex.plugin.on import _PROXY_REGISTRY

from . import _manager, call_plugin_api, get_loaded_plugins


def load_plugins(*plugin_dir: str) -> set[Plugin]:
    return _manager.load_plugins(plugin_dir)


def load_builtin_plugins(*names: str) -> set[Plugin]:
    return load_plugins(*[f"framex.plugins.{name}" for name in names])


@logger.catch(reraise=True)
def auto_load_plugins(builtin_plugins: list[str], plugins: list[str], enable_proxy: bool = False) -> set[Plugin]:
    # Get all builtin_plugins
    loaded_builtin_plugins = {
        plugin.name for plugin in get_loaded_plugins() if plugin.module_name.startswith("framex.plugins.")
    }
    # Candidate builtin_plugins
    candidate_builtin_plugins = set(builtin_plugins) - loaded_builtin_plugins
    # Check if proxy is enabled but not allow to load
    if (
        enable_proxy and "proxy" not in loaded_builtin_plugins and "proxy" not in candidate_builtin_plugins
    ):  # pragma: no cover
        raise RuntimeError(
            "`enable_proxy` == True, but `proxy` is not in `load_builtin_plugins`.\n"
            "Please add `proxy` to `load_builtin_plugins` in your settings file."
        )

    # Load other plugins from settings
    builtin_plugin_instances = load_builtin_plugins(*candidate_builtin_plugins) if candidate_builtin_plugins else set()
    plugin_instances = load_plugins(*plugins) if plugins else set()
    return builtin_plugin_instances | plugin_instances


async def register_proxy_func(func: Callable) -> None:
    from framex.config import settings

    if not settings.server.enable_proxy:  # pragma: no cover
        logger.warning(f"Proxy is not enabled, skip registering proxy function({func.__name__}).")
        return
    full_func_name = f"{func.__module__}.{func.__name__}"
    if full_func_name not in _PROXY_REGISTRY:  # pragma: no cover
        raise RuntimeError(f"Function {full_func_name} is not registered as a proxy function.")

    api_reg = PluginApi(
        deployment_name=PROXY_PLUGIN_NAME,
        call_type=ApiType.PROXY,
        func_name="register_proxy_function",
    )
    await call_plugin_api(
        api_reg,
        None,
        func_name=full_func_name,
        func_callable=_PROXY_REGISTRY[full_func_name],
    )



from collections.abc import Callable
from typing import Any

from pydantic import BaseModel


class ProxyFunc(BaseModel):
    func: Callable[..., Any]
    is_remote: bool = False


class ProxyFuncHttpBody(BaseModel):
    data: str
    func_name: str



from datetime import UTC, datetime, timedelta
from types import SimpleNamespace
from unittest.mock import AsyncMock, Mock, patch
from urllib.parse import urlparse

import jwt
import pytest
from fastapi import HTTPException, status
from fastapi.testclient import TestClient
from starlette.requests import Request

from framex.config import AuthConfig
from framex.consts import DOCS_URL
from framex.driver.application import create_fastapi_application
from framex.driver.auth import auth_jwt, authenticate, create_jwt, oauth_callback

# =========================================================
# helpers
# =========================================================


def fake_oauth(**overrides):
    data = dict(  # noqa: C408
        authorization_url="https://oauth.example.com/authorize",
        token_url="https://oauth.example.com/token",  # noqa: S106
        user_info_url="https://oauth.example.com/user",
        client_id="client",
        client_secret="secret",  # noqa: S106
        redirect_uri="/oauth/callback",
        call_back_url="http://test/callback",
        jwt_secret="secret",  # noqa: S106
        jwt_algorithm="HS256",
    )
    data.update(overrides)
    return SimpleNamespace(**data)


# =========================================================
# create_jwt
# =========================================================


class TestCreateJWT:
    def test_create_jwt_success(self):
        with patch("framex.config.settings.auth.oauth", fake_oauth()):
            token = create_jwt({"username": "test"})
            decoded = jwt.decode(token, "secret", algorithms=["HS256"])
            assert decoded["username"] == "test"
            assert "iat" in decoded
            assert "exp" in decoded

    def test_create_jwt_no_oauth(self):
        with patch("framex.config.settings.auth.oauth", None), pytest.raises(RuntimeError):
            create_jwt({"username": "test"})


# =========================================================
# auth_jwt
# =========================================================


class TestAuthJWT:
    def test_returns_false_when_oauth_not_configured(self):
        with patch("framex.config.settings.auth.oauth", None):
            req = Mock(spec=Request)
            assert auth_jwt(req) is False

    def test_returns_false_when_no_token_cookie(self):
        with patch("framex.config.settings.auth.oauth") as mock_oauth:
            mock_oauth.jwt_secret = "secret"  # noqa: S105
            mock_oauth.jwt_algorithm = "HS256"

            req = Mock(spec=Request)
            req.cookies.get.return_value = None

            assert auth_jwt(req) is False

    def test_returns_true_when_token_is_valid(self):
        with patch("framex.config.settings.auth.oauth") as mock_oauth:
            mock_oauth.jwt_secret = "secret"  # noqa: S105
            mock_oauth.jwt_algorithm = "HS256"

            now = datetime.now(UTC)
            token = jwt.encode(
                {
                    "username": "test",
                    "iat": int(now.timestamp()),
                    "exp": int((now + timedelta(hours=1)).timestamp()),
                },
                "secret",
                algorithm="HS256",
            )

            req = Mock(spec=Request)
            req.cookies.get.return_value = token

            assert auth_jwt(req) is True

    def test_returns_false_when_token_is_invalid(self):
        with patch("framex.config.settings.auth.oauth") as mock_oauth:
            mock_oauth.jwt_secret = "secret"  # noqa: S105
            mock_oauth.jwt_algorithm = "HS256"

            req = Mock(spec=Request)
            req.cookies.get.return_value = "this.is.not.a.jwt"

            assert auth_jwt(req) is False

    def test_returns_false_when_token_is_expired(self):
        with patch("framex.config.settings.auth.oauth") as mock_oauth:
            mock_oauth.jwt_secret = "secret"  # noqa: S105
            mock_oauth.jwt_algorithm = "HS256"

            now = datetime.now(UTC)
            expired_token = jwt.encode(
                {
                    "username": "test",
                    "iat": int((now - timedelta(days=2)).timestamp()),
                    "exp": int((now - timedelta(days=1)).timestamp()),
                },
                "secret",
                algorithm="HS256",
            )

            req = Mock(spec=Request)
            req.cookies.get.return_value = expired_token

            assert auth_jwt(req) is False


# =========================================================
# authenticate (unit)
# =========================================================


class TestAuthenticate:
    def test_redirect_when_no_credentials(self):
        with patch("framex.config.settings.auth.oauth", fake_oauth()):
            req = Mock(spec=Request)
            req.cookies.get.return_value = None
            req.url.path = "/docs"
            with pytest.raises(HTTPException) as exc:
                authenticate(req, api_key=None)
            assert exc.value.status_code == status.HTTP_302_FOUND

    def test_valid_api_key(self):
        with (
            patch("framex.config.settings.auth.oauth", fake_oauth()),
            patch.object(AuthConfig, "get_auth_keys", return_value=["good-key"]),
        ):
            req = Mock(spec=Request)
            req.cookies.get.return_value = None
            req.url.path = "/docs"
            authenticate(req, api_key="good-key")


# =========================================================
# oauth_callback
# =========================================================


class TestOAuthCallback:
    @pytest.mark.asyncio
    async def test_oauth_callback_success(self):
        with patch("framex.config.settings.auth.oauth", fake_oauth()), patch("httpx.AsyncClient") as mock_client_cls:
            client = AsyncMock()
            mock_client_cls.return_value.__aenter__.return_value = client

            token_resp = Mock()
            token_resp.json.return_value = {"access_token": "oauth-token"}
            token_resp.raise_for_status = Mock()

            user_resp = Mock()
            user_resp.json.return_value = {"username": "test", "email": "a@b.com"}
            user_resp.raise_for_status = Mock()

            client.post.return_value = token_resp
            client.get.return_value = user_resp

            res = await oauth_callback(code="abc")
            assert res.status_code == status.HTTP_302_FOUND
            assert res.headers["location"] == DOCS_URL
            assert "token=" in res.headers.get("set-cookie", "")


# =========================================================
# Integration
# =========================================================


class TestAuthenticationIntegration:
    def test_docs_redirects_when_not_authenticated(self):
        with patch("framex.config.settings.auth.oauth", fake_oauth()):
            app = create_fastapi_application()
            client = TestClient(app)

            resp = client.get("/docs", follow_redirects=False)

            assert resp.status_code == status.HTTP_302_FOUND
            location = resp.headers["location"]
            parsed = urlparse(location)
            assert parsed.hostname == "oauth.example.com"

    def test_docs_accessible_with_valid_jwt(self):
        with patch("framex.config.settings.auth.oauth", fake_oauth()):
            app = create_fastapi_application()
            client = TestClient(app)

            now = datetime.now(UTC)
            token = jwt.encode(
                {
                    "username": "test",
                    "iat": int(now.timestamp()),
                    "exp": int((now + timedelta(hours=1)).timestamp()),
                },
                "secret",
                algorithm="HS256",
            )
            client.cookies.set("token", token)
            resp = client.get("/docs", follow_redirects=False)
            assert resp.status_code == status.HTTP_200_OK



import abc
from collections.abc import Callable
from enum import StrEnum
from typing import Any, cast

from aiocache import Cache, cached
from fastapi import FastAPI

from framex.consts import PROXY_PLUGIN_NAME
from framex.plugin.model import ApiType, PluginApi


class AdapterMode(StrEnum):
    LOCAL = "noray"
    RAY = "ray"


class BaseAdapter(abc.ABC):
    mode: AdapterMode

    def to_ingress(self, cls: type, app: FastAPI, **kwargs: Any) -> type:  # noqa: ARG002
        return cls

    def to_deployment(self, cls: type, **kwargs: Any) -> type:  # noqa: ARG002
        return cls

    async def _resolve_stream(self, api: PluginApi, kwargs: dict[str, Any]) -> bool:
        if api.call_type == ApiType.PROXY and api.api:
            kwargs["proxy_path"] = api.api
            return bool(await self._check_is_gen_api(api.api))
        return api.stream

    @abc.abstractmethod
    async def _invoke(self, func: Callable[..., Any], **kwargs: Any) -> Any: ...

    async def call_func(self, api: PluginApi, **kwargs: Any) -> Any:
        func = self.get_handle_func(api.deployment_name, api.func_name)
        stream = await self._resolve_stream(api, kwargs)
        if stream:
            return [chunk async for chunk in self._stream_call(func, **kwargs)]
        return await self._invoke(func, **kwargs)

    def get_handle_func(self, deployment_name: str, func_name: str) -> Any:
        handle = self.get_handle(deployment_name)
        if handle and (func := getattr(handle, func_name)):
            return func
        raise RuntimeError(f"No handle or function found for deployment({deployment_name}:{func_name})")

    @cached(cache=Cache.MEMORY)
    async def _check_is_gen_api(self, path: str) -> bool:
        func = self.get_handle_func(PROXY_PLUGIN_NAME, "check_is_gen_api")
        return cast(bool, await self._acall(func, path=path))

    @abc.abstractmethod
    def get_handle(self, deployment_name: str) -> Any: ...

    @abc.abstractmethod
    def bind(self, deployment: Callable[..., Any], **kwargs: Any) -> Any: ...

    @abc.abstractmethod
    def to_remote_func(self, func: Callable) -> Callable: ...

    @abc.abstractmethod
    def _stream_call(self, func: Callable[..., Any], **kwargs: Any) -> Any: ...

    @abc.abstractmethod
    async def _acall(self, func: Callable[..., Any], **kwargs: Any) -> Any: ...
    @abc.abstractmethod
    def _call(self, func: Callable[..., Any], **kwargs: Any) -> Any: ...



from typing import Any

import framex
from framex.cli import framex as framex_cli
from framex.config import settings
from framex.consts import VERSION


def test_framex_version_option(runner):
    """framex --version prints version"""
    result = runner.invoke(framex_cli, ["--version"])
    assert result.exit_code == 0
    assert VERSION in result.output


def test_run_command_with_defaults(monkeypatch, runner):
    """
    framex run
    - should NOT override settings
    - should call framex.run()
    """
    called = {"called": False}

    def fake_run(*_: Any, **__: Any) -> None:
        called["called"] = True

    monkeypatch.setattr(framex, "run", fake_run)

    original_host = settings.server.host
    original_port = settings.server.port
    original_use_ray = settings.server.use_ray

    result = runner.invoke(framex_cli, ["run"])

    assert result.exit_code == 0
    assert called["called"] is True
    assert "🚀 Starting FrameX with configuration" in result.output

    assert settings.server.host == original_host
    assert settings.server.port == original_port
    assert settings.server.use_ray == original_use_ray


def test_run_command_override_basic_options(monkeypatch, runner) -> None:
    """
    framex run --host --port
    - should override settings
    """

    def fake_run(*args: Any, **kwargs: Any) -> None:
        pass

    monkeypatch.setattr(framex, "run", fake_run)

    result = runner.invoke(
        framex_cli,
        [
            "run",
            "--host",
            "127.0.0.1",
            "--port",
            "9999",
        ],
    )

    assert result.exit_code == 0
    assert settings.server.host == "127.0.0.1"
    assert settings.server.port == 9999


def test_run_command_tristate_bool(monkeypatch, runner) -> None:
    """
    --use-ray / --no-use-ray should override settings
    """

    def fake_run(*args: Any, **kwargs: Any) -> None:
        pass

    monkeypatch.setattr(framex, "run", fake_run)

    result = runner.invoke(framex_cli, ["run", "--no-use-ray"])
    assert result.exit_code == 0
    assert settings.server.use_ray is False

    result = runner.invoke(framex_cli, ["run", "--use-ray"])
    assert result.exit_code == 0
    assert settings.server.use_ray is True


def test_run_command_with_dashboard_options(monkeypatch, runner) -> None:
    """
    Test --dashboard-host and --dashboard-port options
    """

    def fake_run(*args, **kwargs: Any) -> None:
        pass

    monkeypatch.setattr(framex, "run", fake_run)

    result = runner.invoke(
        framex_cli,
        [
            "run",
            "--dashboard-host",
            "0.0.0.0",  # noqa: S104
            "--dashboard-port",
            "8265",
        ],
    )

    assert result.exit_code == 0
    assert settings.server.dashboard_host == "0.0.0.0"  # noqa: S104
    assert settings.server.dashboard_port == 8265


def test_run_command_with_num_cpus(monkeypatch, runner) -> None:
    """
    Test --num-cpus option
    """

    def fake_run(*args: Any, **kwargs: Any) -> None:
        pass

    monkeypatch.setattr(framex, "run", fake_run)

    result = runner.invoke(framex_cli, ["run", "--num-cpus", "4"])

    assert result.exit_code == 0
    assert settings.server.num_cpus == 4


def test_run_command_with_enable_proxy(monkeypatch, runner) -> None:
    """
    Test --enable-proxy / --no-enable-proxy options
    """

    def fake_run(*args: Any, **kwargs: Any) -> None:
        pass

    monkeypatch.setattr(framex, "run", fake_run)

    # Test --enable-proxy
    result = runner.invoke(framex_cli, ["run", "--enable-proxy"])
    assert result.exit_code == 0
    assert settings.server.enable_proxy is True

    # Test --no-enable-proxy
    result = runner.invoke(framex_cli, ["run", "--no-enable-proxy"])
    assert result.exit_code == 0
    assert settings.server.enable_proxy is False


def test_run_command_with_load_plugins(monkeypatch, runner) -> None:
    """
    Test --load-plugins and --load-builtin-plugins options
    """
    loaded_plugins: list[str] = []
    loaded_builtin_plugins: list[str] = []

    def fake_load_plugins(*plugins) -> None:
        loaded_plugins.extend(plugins)

    def fake_load_builtin_plugins(*plugins) -> None:
        loaded_builtin_plugins.extend(plugins)

    def fake_run(*args: Any, **kwargs: Any) -> None:
        pass

    monkeypatch.setattr(framex, "load_plugins", fake_load_plugins)
    monkeypatch.setattr(framex, "load_builtin_plugins", fake_load_builtin_plugins)
    monkeypatch.setattr(framex, "run", fake_run)

    result = runner.invoke(
        framex_cli,
        [
            "run",
            "--load-plugins",
            "plugin1",
            "--load-plugins",
            "plugin2",
            "--load-builtin-plugins",
            "builtin1",
        ],
    )

    assert result.exit_code == 0
    assert "plugin1" in loaded_plugins
    assert "plugin2" in loaded_plugins
    assert "builtin1" in loaded_builtin_plugins



from unittest.mock import Mock, patch

import pytest
from fastapi.routing import APIRoute
from starlette.routing import Route

from framex.driver.ingress import APIIngress

# ---------- helpers ----------


def make_route(path: str, methods: set[str]) -> APIRoute:
    route = Mock(spec=APIRoute)
    route.path = path
    route.methods = methods
    return route


# ---------- fixtures ----------


@pytest.fixture
def mock_app():
    with patch("framex.driver.ingress.app") as app:
        app.routes = []
        app.add_api_route = Mock()
        yield app


@pytest.fixture
def ingress():
    return APIIngress.__new__(APIIngress)


# ---------- tests ----------


def test_add_first_route_success(ingress, mock_app):
    endpoint = Mock()

    ingress.add_api_route("/users", endpoint, methods=["GET"])

    mock_app.add_api_route.assert_called_once_with("/users", endpoint, methods=["GET"])


@pytest.mark.parametrize(
    ("existing_path", "new_path"),
    [
        ("/users/{id}", "/users/{id}"),
        ("/users/{id}", "/users/{user_id}"),
        ("/users/{uid}/posts/{pid}", "/users/{id}/posts/{post_id}"),
    ],
)
def test_duplicate_path_same_method_raises(ingress, mock_app, existing_path, new_path):
    mock_app.routes = [make_route(existing_path, {"GET"})]

    with pytest.raises(RuntimeError, match=r"Duplicate API route"):
        ingress.add_api_route(new_path, Mock(), methods=["GET"])


def test_same_path_different_method_allowed(ingress, mock_app):
    mock_app.routes = [make_route("/users/{id}", {"GET"})]

    ingress.add_api_route("/users/{id}", Mock(), methods=["POST"])

    mock_app.add_api_route.assert_called_once()


def test_overlapping_methods_raises(ingress, mock_app):
    mock_app.routes = [make_route("/users", {"GET", "POST"})]

    with pytest.raises(RuntimeError):
        ingress.add_api_route("/users", Mock(), methods=["POST", "PUT"])


def test_case_insensitive_methods(ingress, mock_app):
    mock_app.routes = [make_route("/users", {"GET"})]

    with pytest.raises(RuntimeError):
        ingress.add_api_route("/users", Mock(), methods=["get"])


def test_non_api_route_is_ignored(ingress, mock_app):
    non_api_route = Mock(spec=Route)
    non_api_route.path = "/users/{id}"
    mock_app.routes = [non_api_route]

    ingress.add_api_route("/users/{id}", Mock(), methods=["GET"])

    mock_app.add_api_route.assert_called_once()


def test_kwargs_are_passed_through(ingress, mock_app):
    ingress.add_api_route(
        "/users",
        Mock(),
        methods=["GET"],
        tags=["users"],
        response_class=Mock(),
    )

    _, kwargs = mock_app.add_api_route.call_args
    assert kwargs["tags"] == ["users"]
    assert "response_class" in kwargs



from datetime import UTC, datetime, timedelta

import httpx
import jwt
from fastapi import Depends, HTTPException, Response, status
from fastapi.responses import RedirectResponse
from fastapi.security import APIKeyHeader
from starlette.requests import Request

from framex.config import settings
from framex.consts import DOCS_URL

api_key_header = APIKeyHeader(name="Authorization", auto_error=False)


def create_jwt(payload: dict) -> str:
    if not settings.auth.oauth:
        raise RuntimeError("OAuth not configured")

    now_utc = datetime.now(UTC)

    payload.update(
        {
            "iat": int(now_utc.timestamp()),
            "exp": int((now_utc + timedelta(hours=24)).timestamp()),
        }
    )
    return jwt.encode(payload, settings.auth.oauth.jwt_secret, algorithm=settings.auth.oauth.jwt_algorithm)


def auth_jwt(request: Request) -> bool:
    if not settings.auth.oauth:
        return False

    token = request.cookies.get("token")
    if not token:
        return False

    try:
        jwt.decode(
            token,
            settings.auth.oauth.jwt_secret,
            algorithms=[settings.auth.oauth.jwt_algorithm],
        )
        return True
    except (jwt.InvalidTokenError, jwt.ExpiredSignatureError):
        return False


def authenticate(request: Request, api_key: str | None = Depends(api_key_header)) -> None:
    if settings.auth.oauth:
        if token := request.cookies.get("token"):
            try:
                jwt.decode(
                    token,
                    settings.auth.oauth.jwt_secret,
                    algorithms=[settings.auth.oauth.jwt_algorithm],
                )
                return

            except Exception as e:
                from framex.log import logger

                logger.warning(f"JWT decode failed: {e}")

        if api_key and api_key in (settings.auth.get_auth_keys(request.url.path) or []):
            return

        raise HTTPException(
            status_code=status.HTTP_302_FOUND,
            headers={
                "Location": (
                    f"{settings.auth.oauth.authorization_url}"
                    f"?client_id={settings.auth.oauth.client_id}"
                    "&response_type=code"
                    f"&redirect_uri={settings.auth.oauth.call_back_url}"
                    "&scope=read_user"
                )
            },
        )


async def oauth_callback(code: str) -> Response:
    if not settings.auth.oauth:  # pragma: no cover
        raise RuntimeError("OAuth not configured")

    async with httpx.AsyncClient() as client:
        resp = await client.post(
            settings.auth.oauth.token_url,
            data={
                "client_id": settings.auth.oauth.client_id,
                "client_secret": settings.auth.oauth.client_secret,
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": settings.auth.oauth.call_back_url,
            },
        )
        resp.raise_for_status()
        token_resp = resp.json()

        if not (auth_token := token_resp.get("access_token")):  # pragma: no cover
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="GitLab token exchange failed")

        resp = await client.get(
            settings.auth.oauth.user_info_url,
            headers={"Authorization": f"Bearer {auth_token}"},
            timeout=10,
        )
        resp.raise_for_status()
        user_resp = resp.json()

    if not (username := user_resp.get("username")):  # pragma: no cover
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Failed to get GitLab user")

    user_info = {
        "message": f"Welcome {username}",
        "username": username,
        "email": user_resp.get("email"),
    }

    res = RedirectResponse(url=DOCS_URL, status_code=status.HTTP_302_FOUND)
    res.set_cookie(
        "token",
        create_jwt(user_info),
        httponly=True,
        samesite="lax",
    )
    return res



import click

from framex.consts import VERSION


@click.group()
@click.version_option(VERSION, prog_name="framex")
def framex() -> None:
    """FrameX command line interface."""


@framex.command()
@click.option(
    "--host",
    default=None,
    help="Host address to bind the FrameX service.",
)
@click.option(
    "--port",
    default=None,
    type=int,
    help="Port number for the FrameX service.",
)
@click.option(
    "--dashboard-host",
    default=None,
    help="Host address for the Ray Dashboard.",
)
@click.option(
    "--dashboard-port",
    default=None,
    type=int,
    help="Port number for the Ray Dashboard.",
)
@click.option(
    "--num-cpus",
    default=None,
    type=int,
    help="Number of CPU cores allocated to Ray.",
)
@click.option("--load-plugins", multiple=True, help="List of external plugins to load. Can be used multiple times.")
@click.option(
    "--load-builtin-plugins", multiple=True, help="List of built-in plugins to load. Can be used multiple times."
)
@click.option(
    "--use-ray/--no-use-ray",
    default=None,
    help="Enable or disable Ray. If not set, use config value.",
)
@click.option(
    "--enable-proxy/--no-enable-proxy",
    default=None,
    help="Enable or disable HTTP proxy. If not set, use config value.",
)
def run(
    host: str | None,
    port: int | None,
    dashboard_host: str | None,
    dashboard_port: int | None,
    num_cpus: int | None,
    load_plugins: tuple[str, ...],
    load_builtin_plugins: tuple[str, ...],
    use_ray: bool | None,
    enable_proxy: bool | None,
) -> None:
    """Run the FrameX service."""
    import framex as fx
    from framex.config import settings

    # Create an updated config instance
    if host is not None:
        settings.server.host = host
    if port is not None:
        settings.server.port = port
    if dashboard_host is not None:
        settings.server.dashboard_host = dashboard_host
    if dashboard_port is not None:
        settings.server.dashboard_port = dashboard_port
    if num_cpus is not None:
        settings.server.num_cpus = num_cpus
    if use_ray is not None:
        settings.server.use_ray = use_ray
    if enable_proxy is not None:
        settings.server.enable_proxy = enable_proxy

    click.echo("🚀 Starting FrameX with configuration:")

    fx.load_plugins(*load_plugins)
    fx.load_builtin_plugins(*load_builtin_plugins)

    fx.run()



"""Tests for framex.adapter.__init__ module."""

from unittest.mock import MagicMock, patch

from framex.adapter import get_adapter
from framex.adapter.base import BaseAdapter
from framex.adapter.local_adapter import LocalAdapter


class TestGetAdapter:
    """Tests for the get_adapter factory function."""

    def setup_method(self):
        """Reset the global adapter before each test."""
        import framex.adapter as adapter_module

        adapter_module._adapter = None

    def test_get_adapter_returns_local_adapter_when_ray_disabled(self):
        """Test get_adapter returns LocalAdapter when use_ray is False."""
        with patch("framex.adapter.settings.server.use_ray", False):
            adapter = get_adapter()
            assert isinstance(adapter, LocalAdapter)
            assert isinstance(adapter, BaseAdapter)

    def test_get_adapter_returns_ray_adapter_when_ray_enabled(self):
        """Test get_adapter returns RayAdapter when use_ray is True."""
        with (
            patch("framex.adapter.settings.server.use_ray", True),
            patch("framex.adapter.ray_adapter.RayAdapter") as mock_ray_adapter,
        ):
            mock_instance = MagicMock()
            mock_ray_adapter.return_value = mock_instance

            adapter = get_adapter()
            assert adapter == mock_instance
            mock_ray_adapter.assert_called_once()

    def test_get_adapter_returns_same_instance_on_multiple_calls(self):
        """Test get_adapter returns the same singleton instance."""
        with patch("framex.adapter.settings.server.use_ray", False):
            adapter1 = get_adapter()
            adapter2 = get_adapter()
            assert adapter1 is adapter2

    def test_get_adapter_caches_local_adapter(self):
        """Test that LocalAdapter is cached after first call."""
        with patch("framex.adapter.settings.server.use_ray", False):
            adapter1 = get_adapter()
            # Second call should return cached instance
            adapter2 = get_adapter()
            assert adapter1 is adapter2
            assert isinstance(adapter1, LocalAdapter)

    def test_get_adapter_caches_ray_adapter(self):
        """Test that RayAdapter is cached after first call."""
        with (
            patch("framex.adapter.settings.server.use_ray", True),
            patch("framex.adapter.ray_adapter.RayAdapter") as mock_ray_adapter,
        ):
            mock_instance = MagicMock()
            mock_ray_adapter.return_value = mock_instance

            adapter1 = get_adapter()
            adapter2 = get_adapter()

            # Should only instantiate once
            assert adapter1 is adapter2
            assert mock_ray_adapter.call_count == 1

    def test_get_adapter_lazy_imports_ray_adapter(self):
        """Test that RayAdapter is only imported when needed."""
        with patch("framex.adapter.settings.server.use_ray", False):  # noqa
            # Import should not happen when use_ray is False
            with patch("framex.adapter.ray_adapter") as mock_ray_module:
                get_adapter()
                # RayAdapter module should not be accessed
                mock_ray_module.RayAdapter.assert_not_called()

    def test_adapter_initially_none(self):
        """Test that _adapter global is None before first call."""
        import framex.adapter as adapter_module

        adapter_module._adapter = None
        assert adapter_module._adapter is None

    def test_adapter_set_after_first_call(self):
        """Test that _adapter global is set after first call."""
        import framex.adapter as adapter_module

        adapter_module._adapter = None
        with patch("framex.adapter.settings.server.use_ray", False):
            get_adapter()
            assert adapter_module._adapter is not None
            assert isinstance(adapter_module._adapter, LocalAdapter)

    def test_get_adapter_with_switching_ray_setting(self):
        """Test that once adapter is set, changing ray setting doesn't affect it."""
        import framex.adapter as adapter_module

        adapter_module._adapter = None

        with patch("framex.adapter.settings.server.use_ray", False):
            adapter1 = get_adapter()
            assert isinstance(adapter1, LocalAdapter)

        # Change setting, but adapter should still be the same
        with patch("framex.adapter.settings.server.use_ray", True):
            adapter2 = get_adapter()
            # Should still be the same LocalAdapter instance
            assert adapter2 is adapter1
            assert isinstance(adapter2, LocalAdapter)



import secrets
from typing import Any, Literal

from pydantic import BaseModel, Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    PyprojectTomlConfigSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)


class LogConfig(BaseModel):
    simple_log: bool = True
    ignored_prefixes: tuple[str, ...] = (
        "Started executing request to method",
        "CALL register_route",
        "CALL /api/v1",
        "Initialized DeploymentHandle",
        "Finished initializing replica",
        "Got updated replicas for",
        "Started <ray.serve._private.router.SharedRouterLongPollClient object",
    )
    ignored_contains: tuple[str, ...] = ("GET /ping", "GET /health")


class SentryConfig(BaseModel):
    enable: bool = False
    dsn: str = ""
    env: str = ""  # local, prod, dev
    debug: bool = False
    ignore_errors: list[str] = Field(default_factory=list)
    lifecycle: Literal["manual", "trace"] = "manual"
    enable_logs: bool = False


class ServerConfig(BaseModel):
    host: str = "127.0.0.1"
    port: int = 8080
    dashboard_host: str = "127.0.0.1"
    dashboard_port: int = 8260
    use_ray: bool = False
    enable_proxy: bool = False
    legal_proxy_code: list[int] = Field(default_factory=lambda: [200])
    num_cpus: int = -1
    excluded_log_paths: list[str] = Field(default_factory=list)
    ingress_config: dict[str, Any] = Field(default_factory=lambda: {"max_ongoing_requests": 60})
    reversion: str = ""


class TestConfig(BaseModel):
    disable_record_request: bool = False
    silent: bool = False


class OauthConfig(BaseModel):
    client_id: str = ""
    client_secret: str = ""
    authorization_url: str = ""
    redirect_uri: str = ""
    base_url: str = ""

    token_url: str = ""
    user_info_url: str = ""
    app_url: str = ""

    jwt_secret: str = ""
    jwt_algorithm: str = "HS256"

    @property
    def call_back_url(self) -> str:
        return f"{self.app_url}{self.redirect_uri}"

    def model_post_init(self, context: Any) -> None:
        super().model_post_init(context)
        if not self.authorization_url:
            self.authorization_url = f"{self.base_url}/oauth/authorize"
        if not self.token_url:
            self.token_url = f"{self.base_url}/oauth/token"
        if not self.user_info_url:
            self.user_info_url = f"{self.base_url}/api/v4/user"
        if not self.jwt_secret:
            self.jwt_secret = secrets.token_urlsafe(32)


class AuthConfig(BaseModel):
    oauth: OauthConfig | None = Field(default=None)
    rules: dict[str, list[str]] = Field(default_factory=dict)

    def _is_url_protected(self, url: str) -> bool:
        for rule in self.rules:
            if rule == url:
                return True
            if rule.endswith("/*") and url.startswith(rule[:-1]):
                return True
        return False

    def get_auth_keys(self, url: str) -> list[str] | None:
        if not self._is_url_protected(url):
            return None

        if url in self.rules:
            return self.rules[url]

        matched_keys = None
        matched_len = -1

        for rule, keys in self.rules.items():
            if not rule.endswith("/*"):
                continue

            prefix = rule[:-1]
            if url.startswith(prefix) and len(prefix) > matched_len:
                matched_keys = keys
                matched_len = len(prefix)

        return matched_keys


class Settings(BaseSettings):
    # Global config
    base_ingress_config: dict[str, Any] = Field(default_factory=lambda: {"max_ongoing_requests": 10})

    server: ServerConfig = Field(default_factory=ServerConfig)
    log: LogConfig = Field(default_factory=LogConfig)

    # plugins config
    plugins: dict[str, Any] = Field(default_factory=dict)
    load_plugins: list[str] = Field(default_factory=list)
    load_builtin_plugins: list[str] = Field(default_factory=list)

    test: TestConfig = Field(default_factory=TestConfig)
    sentry: SentryConfig = Field(default_factory=SentryConfig)
    auth: AuthConfig = Field(default_factory=AuthConfig)

    model_config = SettingsConfigDict(
        # `.env.prod` takes priority over `.env`
        env_file=(".env", ".env.prod"),
        extra="ignore",
        env_nested_delimiter="__",
        case_sensitive=False,
        pyproject_toml_table_header=("tool", "framex"),
        toml_file="config.toml",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            env_settings,
            dotenv_settings,
            TomlConfigSettingsSource(settings_cls),
            PyprojectTomlConfigSettingsSource(settings_cls),
            init_settings,
            file_secret_settings,
        )


settings = Settings()
